<?php

/* 

PHPallus: PHP abstraction library for learning, utilities and services

Copyright (C) 2001 Emmanuel Chazard http://emmanuel.chazard.org
Check for updates on http://phpallus.chazard.org

This program is free software: you can redistribute it and/or modify
it under the terms of the CeCILL license, version 2.1 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
CeCILL License for more details.

You should have received a copy of the CeCILL License
along with this program.  If not, see <http://www.cecill.info/licences.fr.html>.




Version 3.3
Changelog : 

- 2022-08-03 : ajout de la validation de type VALIDATE_EMAIL_UNIV_LILLE dans la classe chaz_field (compatibilité assurée)
	Ajout de la fonction autoselect(), mise à jour de autocomplete()
	Modification de la fonction check_unique() (compatibilité assurée)
	Modification de la fonction suggestion_lien()  (compatibilité assurée)
	Ajout de la classe chaz_field_file qui permet d'uploader des fichiers (non sérialisable en form)
	Dans chaz_entity, ajout de la propriété form_encoding pour gérer les formulaires avec upload.
	Révision complète de la fonction chaz_rename_file
	Retrait du typage fort du paramètre de la fonction chaz_field::set_value_from_unreliable_source()
	Ajout de la fonction file_download()
- 2023-03-09 correction du constructeur de chaz_field_file::_construct
- 2023-03-28 : ajout de la classe chaz_field_file2 qui gère les champs file, 
	mais par un processus distinct (autre table ou fichier physique par exemple)
- 2023-08-02 : ajout de la classe chaz_entity_form qui générique la mécanique de sérialisation 
	des champs "form:*" dans un blob unique "form". 
	La méthode get_additional_fieldnames() de chaz_entity y est déplacée.
	Ajout de la fonction expandable_label_for_listing() pour afficher du texte long dans un listing.
	Dans chaz_field, Correction des procédés TRANSFORM_TONUMERIC TRANSFORM_TOINTEGER TRANSFORM_TOPOSITIVEINTEGER qui étaient erronés
	Ajout de chaz_entity_repository::db_select_many_items_with_query_indexed() qui indexe le résultat par ID
- 2023-09-15 : ajout de chaz_field::update_values() qui permet de recalculer les valeurs affichées d'un champ, 
	utile lorsqu'on modifie a posteriori les options d'une liste déroulante par exemple
- 2023-09-21 : ajout de la classe chaz_field_tags qui permet d'utiliser Jquery Tagit (tags)
	Ajout des procédés de transformation TRANSFORM_DEDUPLICATE et TRANSFORM_SORT et TRANSFORM_TAGIT dans chaz_field
	Penser à avoir un style CSS .tagit_statique
- 2023-09-26 modification de chaz_entity::__get : désormais, on peut appeler la version SQL d'un champ en utilisant 
	son nom terminé par underscore (ex : $user->sexe_ au lieu de $user->sexe). En outre,
	si le champ n'est pas trouvé, cette méthode tente de le trouver en replaçant "form_" par "form:" en début de champ.
	Ex : $user->form_preference est lu comme $user->get_field_value_html('form:preference') si ce champ existe
	et   $user->form_preference_ est lu comme $user->get_field_value_sql('form:preference') si ce champ existe
- 2023-09-28 fonction new_checkbox pour faire des checkbox qui se comportent comme des radiobox binaires 
- 2023-10-01 fonction jalon pour afficher une barre de texte avec un numéro auto-incrémenté
- 2023-10-02 fonction licence()

TODO : 
- composant de formulaire hybride : texte libre avec en plus une liste déroulante d'options possibles, mais pas limitante
- composant EVA (radiobox mises en forme)
- déclinaison de chaz_field_file2 pour les photos d'identité
- chaz_field_file2 permettant d'ajouter un fichier déjà existant



*/

// cette fonction propose un radiobox 0/1 (donc toujours présent dans le POST) 
// avec un aspect de sélecteur bouton coulissant oui/non (aspect=oui) ou de case à cocher (aspect=check)
// le gain majeur, outre l'esthétique, est que l'item est toujours présent dans le POST,
// tout comme une radiobox, mais pas comme une checkbox.
function new_checkbox($name, $value=0, $aspect="oui"): string {
	if( $aspect=="oui") {
		$src_oui = "images/checkbox_oui.png" ;
		$src_non = "images/checkbox_non.png" ;
	} elseif( $aspect=="check") {
		$src_oui = "images/checkbox_check.png" ;
		$src_non = "images/checkbox_uncheck.png" ;
	} else {
		trigger_error("Valeurs oui ou check autorisées seulement", E_USER_ERROR) ;
	}
	if( $value==1) {
		$src = $src_oui ;
		$checked_oui = 'checked' ;
		$checked_non = '' ;
	} elseif( $value==0) {
		$src = $src_non ;
		$checked_non = 'checked' ;
		$checked_oui = '' ;
	} else {
		trigger_error("Valeurs 0 ou 1 autorisées seulement", E_USER_NOTICE) ;
	}
	$id = genere_string_aleatoire() ;
	$id_oui = $id.'_oui' ;
	$id_non = $id.'_non' ;
$str = "<img src='$src' onClick=\"
element_oui = document.getElementById('$id_oui') ;
element_non = document.getElementById('$id_non') ;
if( element_oui.checked ) {
	this.src='$src_non' ; 
	element_non.checked=true ; 
	this.alt='non' ;
} else {
	this.src='$src_oui' ; 
	element_oui.checked=true ; 
	this.alt='oui' ;
}  \" />
<span style='display:none;'>
<input type='radio' value='0' name='$name' id='$id_non' $checked_non />0
<input type='radio' value='1' name='$name' id='$id_oui' $checked_oui />1
</span>" ;
	return $str ;
}


// retourne le code HTML d'un bandeau dont le numéro s'auto-incrémente
function jalon($texte_html): string {
	static $compteur = 1 ;
    // temporaire, mais on va l'améliorer !
	$str = "<br /><table cellspacing=0 cellpadding=0 border=0 width=100%>
		<tr height=53>
			<td width=50 valign=middle align=center style=background-image:url('./images/jalon_numero.png');font-size:24px;font-weight:bold;color:white;>$compteur</td>
			<td valign=middle align=left style=background-image:url('./images/jalon_bandeau.png');font-size:16px;font-weight:bold;color:white;>&nbsp;&nbsp;&nbsp;$texte_html</td>
			<td width=20 style=background-image:url('./images/jalon_fin.png');>&nbsp;</td>
		</tr></table>" ;
	$compteur++;
	return $str ;
}

// pour invoquer cette fonction, s'assurer qu'il n'y a aucun PRINT préalable (sauf messages d'erreurs bloquantes)
// On peut ajouter des print par la suite, mais leur contenu fera alors partie du fichier.
function file_download($content, $name="document.txt") {
	$size = strlen($content) ; // vérifié : fonctionne pour binaire car compte directement des octets.
	
	header('Content-Description: File Transfer');
	header('Content-Type: application/octet-stream');
	header("Cache-Control: no-cache, must-revalidate");
	header("Expires: 0");
	header('Content-Disposition: attachment; filename="'.$name.'"');
	header('Content-Length: ' . $size);
	header('Pragma: public');
	
	// flush();

	echo $content ;
	exit ;
}
// génère une chaîne de caractère utile pour les dépendances, la licence, etc.
function licence(): string {
	$str = "<p>Cette plateforme a été développée par le Pr Emmanuel Chazard. Elle est sous licence CeCILL. 
			Elle s'appuie sur différentes librairies open sources :</p>
			<ul>
				<li>PHPallus, sous licence CeCILL (framework PHP d'ORM et autres)</li>
				<li>Securimage, sous licence BSD (librairie PHP générant des Captcha)</li>
				<li>DomPDF, sous licence GNU LGPL (librairie PHP pour la génération de PDF)</li>
				<li>JQuery, sous licence MIT (extension du langage JavaScript)</li>
				<li>JQuery UI, sous licence Creative Commons (extension du langage JavaScript)</li>
				<li>Nicedit, sous licence MIT (éditeur de texte JavaScript)</li>
				<li>Tagit, sous licence MIT (apposition de tags en JavaScript)</li>
				<li>DataTable, sous licence MIT (tableaux dynamiques en JavaScript)</li>
			</ul>" ;
	return $str ;
}

// le masque ne contient pas de dossier cette fois-ci
// le dossier peut être non fourni dans un premier temps, il doit se terminer par un slash
function monglob_recursif( $masque , $path = "./"  ) {
	$files = array() ;
	$masques = preg_split("|[\s,;]+|i", $masque);
	foreach( $masques as $un_masque) {
		$files = array_merge($files, glob($path . $un_masque) ) ;
	}	
	$folders = glob($path."*",GLOB_ONLYDIR) ;
	for($i=0;$i<count($folders);$i++) {
		$files = array_merge( $files, monglob_recursif( $masque , $path . basename($folders[$i]) . '/' )) ;
	}
	$files = array_unique( array_values($files )) ;
	sort( $files ) ;
	return $files ;
}

function chaz_rename_file($nom_fichier, $lowercase=TRUE)   {
	$utf8 = array(
        '/[áàâãªä]/u'   =>   'a',
        '/[ÁÀÂÃÄ]/u'    =>   'A',
        '/[ÍÌÎÏ]/u'     =>   'I',
        '/[íìîï]/u'     =>   'i',
        '/[éèêë]/u'     =>   'e',
        '/[ÉÈÊË]/u'     =>   'E',
        '/[óòôõºö]/u'   =>   'o',
        '/[ÓÒÔÕÖ]/u'    =>   'O',
        '/[úùûü]/u'     =>   'u',
        '/[ÚÙÛÜ]/u'     =>   'U',
        '/ç/'           =>   'c',
        '/Ç/'           =>   'C',
        '/ñ/'           =>   'n',
        '/Ñ/'           =>   'N',
		'/œ/'           =>   'oe',
		'/Œ/'           =>   'OE',
		'/Æ/'           =>   'AE',
		'/æ/'           =>   'ae'
    );
    $nom_fichier = preg_replace(array_keys($utf8), array_values($utf8), $nom_fichier);
	$nom_fichier = preg_replace('/[^0-9a-zA-Z\.]/', '_', $nom_fichier);
	$nom_fichier = preg_replace('/_+/', '_', $nom_fichier);
	if( $lowercase ) {
		$nom_fichier = strtolower($nom_fichier) ;
	}
	return $nom_fichier ;
}





/*
Affichage direct d'un résultat de requête sous forme de tableau HTML triable avec DataTable, 
sans aucune traduction des entêtes ni des valeurs.
*/

function affiche_requete( PDOStatement $statement ) {  
	$nom_col = array() ;
	
	$id_table = 'table'.rand() ;

	$chaine =  "<table class='dataTable display compact' id='$id_table'>" ;
	
	while( $ligne = $statement->fetch( 2 )) {
		// affichage des entêtes
		if( count($nom_col) <1 ) {
			foreach( $ligne as $clef=>$valeur ) {
				// $nom_col[] = libelles_champs::get_lib_html( $clef ) ;
				$nom_col[] =$clef ;
			}
			$chaine .= "<thead><tr><td valign=top class=form_clef>". implode('</td><td valign=top class=form_clef>',$nom_col) ."</td></tr></thead>\r\n<tbody>" ;
		}
		// ici transformation ad hoc de certains champs
		foreach( $ligne as $clef=>$valeur ) {
			if( trim($valeur) !="") {
				// rien (adapté du site des ISP...
			} else {
				$ligne[$clef]  = '&nbsp;'  ;
			}
		}
		// affichage de la ligne
		$chaine .= "<tr><td class=form_val>". implode('</td><td class=form_val>',$ligne) ."</td></tr>\r\n" ;
	}
	$chaine .= "</tbody></table>" ;
	$chaine .= "<script  type='text/javascript' >$(document).ready(function() {  $('#$id_table').DataTable({
		'pageLength': -1,  
		'lengthMenu': [ [25, 50, 100, 200, -1], [25, 50, 100, 200, 'Tous'] ] ,
		dom: 'Bfrtip',
		buttons: [ 
			'copy', 
			{extend: 'csvHtml5',title: 'export'}, 
			{extend: 'excelHtml5',title: 'export'},
			{extend: 'pdfHtml5',title: 'export'}, 
			'print'
		]
		});  } );</script>" ;
	// remplace buttons: [ 'copy', 'csv', 'excel', 'pdf', 'print' ]
	return $chaine ;
}





/*
* Fonction qui permet d'imprimer un script JavaScript qui envoit un message lors de la modification du formulaire,
* si la valeur du champ est déjà présente en base. ATTENTION : cette fonction révèle DANS LE CODE HTML les valeurs
* actuellement disponibles. Elle ne doit donc être utilisée que pour des pages ouvertes aux administrateurs.
*
*  $id_champ est l'identifiant (id=...) du champ de formulaire
*  $requete est une requête SQL qui retourne deux colonnes : une première avec la valeur à chercher, une deuxième avec la manière de la présenter
*           (exemple : "nom" dans le premier cas, et concat(prenom,' ',nom) dans le deuxième cas)
*           ce qui permet par exemple d'avoir un affichage plus sympa, ou un lien hypertexte.
*           Si la requête ne retourne qu'une colonne, elle est réutilisée
*  $nom_champ_joli permet de constituer une phrase du type "la valeur saisie dans le champ xxx est déjà affectée"
*  
*  A noter que les valeurs seront comparées après mise automatique en minuscules
* 
*  Le nom de variable utilisé en javascript est unique, pour éviter les téléscopage en cas d'appels multiples
*  sur la même page.
*/
function check_unique($id_champ, $requete, $binding_array=array(), $nom_champ_joli='précédent') {
	$alea = genere_string_aleatoire(6) ;
	$bdd = chaz_db::getInstance() ;

	$stt = $bdd->prepare($requete) ;
	$stt->execute($binding_array) ;
	$liste_items=array() ;
	$liste_labels=array() ;
	while($tab=$stt->fetch(PDO::FETCH_NUM)) {
		if( strpos($tab[0], '"')===FALSE and strpos($tab[0], "'")===FALSE) {
			$liste_items[]=strtolower($tab[0]) ;
			if( isset($tab[1])) {
				$liste_labels[]=str_replace('"',' ', str_replace("'",' ', $tab[1])) ;
			} else {
				$liste_labels[]=$tab[0] ;
			}
		}
	}
	if( count($liste_items)>0) {
		return "<script language='JavaScript'>
				var checkuniqueitems".$alea." = new Array( \"". implode("\", \"", $liste_items)  ."\") ;
				var checkuniquelabels".$alea." = new Array( \"". implode("\", \"", $liste_labels)  ."\") ;
				var checkuniqueelement".$alea." = document.getElementById('$id_champ');
				checkuniqueelement".$alea.".onblur = function() {
				var checkuniqueindex".$alea." = checkuniqueitems".$alea.".indexOf(checkuniqueelement".$alea.".value.toLowerCase()) ;
				if ( checkuniqueindex".$alea." !== -1) {
				alert('La valeur que vous venez de saisir dans le champ $nom_champ_joli existe déjà ('+checkuniqueelement".$alea.".value+' => '+checkuniquelabels".$alea."[checkuniqueindex".$alea."]+'). S\'il s\'agit du même élément, nous vous conseillons de modifier l\'élément existant plutôt que d\'en créer un nouveau.' );
	}
	}
	</script>" ;
	}
	return "" ;
}



/*
* Fonction qui permet d'imprimer un script JavaScript qui remplit automatiquement un champ si ce champ est vide.
* Imaginé par exemple pour le mot de passe lors de la création d'un utilisateur, car difficile de modifier la classe chaz_field_password
*
*  $id_champ est l'identifiant (id=...) du champ de formulaire
*  $value est la valeur à mettre si vide
*/
function autocomplete($id_champ, $value) {
	$value = str_replace("'"," ", $value) ;
	return "<script language='JavaScript'>
			var autocompletevalue = '$value' ;
			var autocompleteelement = document.getElementById('$id_champ');
			if( autocompleteelement.value == '' ) {  autocompleteelement.value = autocompletevalue ; }
	</script>" ;
	
}

/*
* Fonction qui permet d'imprimer un script JavaScript qui sélectionne automatiquement une option dans un champ select.
* Imaginé par exemple pour forcer contextuellement une valeur par défaut
*
*  $id_champ est l'identifiant (id=...) du champ de formulaire
*  $value est la valeur à mettre si vide
*/
function autoselect($id_champ, $value) {
	$value = str_replace("'"," ", $value) ;
	return "<script language='JavaScript'>
			var autocompletevalue = '$value' ;
			var autocompleteelement = document.getElementById('$id_champ');
			autocompleteelement.value = autocompletevalue ;
	</script>" ;
	
}



// Cette fonction permet d'écrire rapidement un texte suggérant que faire après une action
// (par exemple, après l'ajout d'un utilisateur, gérer les utilisateurs)
// La phrase doit contenir les signes [ et ] pour délimiter un lien hypertexte. Sinon, un " : cliquez ici" sera ajouté à la fin
// Paramètre : un ARRAY avec pour chaque élément lien_http => phrase
// Paramètre check : vérifie si ce fichier existe, en supposant qu'on parte de la racine du site, et après éviction des paramètres GET.
function suggestion_liens(array $items, $check=TRUE) {
	
	$phrases = array() ;
	foreach( $items as $lien => $phrase) {
		
		if( (!$check) OR file_exists($lien) OR file_exists(parse_url($lien, PHP_URL_PATH))) {
			if( strstr($phrase, '[') AND strstr($phrase, ']')) {
				$phrase = str_replace('[', "<a href=$lien>", $phrase) ;
				$phrase = str_replace(']', "</a>", $phrase) ;
			} else {
				$phrase = $phrase."&nbsp;: <a href=$lien>cliquez ici</a>" ;
			}
			$phrases[] = $phrase ;
		} else {
			$trace = debug_backtrace();
			trigger_error("Le lien suivant n'existe pas et ne sera donc pas affiché : $lien".' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'], E_USER_NOTICE) ; 
		}
	}
	// ensuite couche affichage
	
	if( count($phrases)==1) {
		$html = "<div class=whatnext><p><img src=images/ico_whatnext.png width=16 height=16 />Cette page peut vous intéresser : " ;
		$html .= $phrases[0] ."</p></div>" ;
	} else {
		$html = "<div class=whatnext><p><img src=images/ico_whatnext.png width=16 height=16 />Ces pages peuvent vous intéresser : " ;
		$html .= "</p><ul><li>" ;
		$html .= implode("</li><li>", $phrases) ;
		$html .= "</li></ul></div><br />" ;
	}
	return $html ;
}




function genere_string_aleatoire($length=16) {
	$lettres = range('a', 'z') ;
	$str = '' ; 
	for( $i=0; $i<$length; $i++) {
		$offset = rand( 0 , 25 ) ;
		$str .= $lettres[$offset] ;
	}
	return $str ;
}


// cette fonction imprime le code HTML qui permet de cacher le menu, par exemple pour une popup...
// paramètre : print si TRUE, imprime directement, si FALSE, renvoit la chaîne à imprimer.
function cache_menu($print=TRUE) {
	$string = "<STYLE type='text/css'>
			#menu_gauche {
				display: none;
			}
			body {
				margin-left: 5;
				margin-top: 5;
			}
			</STYLE>" ;
	if( $print ) {
		print $string ;
		return TRUE ;
	} else {
		return $string ;
	}
}
// retourne une taille en octets, après conversion automatique en o, ko, Mo, Go en string, comprenant l'unité
// Exemple : pretty_file_size(1757) => "1,8 ko"
function pretty_file_size($size_octets) {
	$size_octets = (int) $size_octets ;
	if( $size_octets >= 1e9 ) {
		$nb = round($size_octets / 1e9,1) ;
		return $nb . ' Go' ;
	} elseif( $size_octets >= 1e6 ) {
		$nb = round($size_octets / 1e6,1) ;
		return $nb . ' Mo' ;
	} elseif( $size_octets >= 1e3 ) {
		$nb = round($size_octets / 1e3,1) ;
		return $nb . ' ko' ;
	} else {
		return $size_octets . ' o' ;
	}
}



/* Propose un rectangle jaune monoligne, idéal pour les listings.
   Lorsqu'on clique dessus, le rectangle occupe autant de lignes que nécessaire.
   Idéal pour des commentaires, ou un log, à faire tenir dans un listing.
*/
function expandable_label_for_listing($txt, $height=20, $style=" background-color:#FFFFB1; border: 1px solid #D4D4D4; ") {
		if( $txt=='') {
			return "" ;
		}
		$txt = str_replace('"',' ',$txt) ; 
		$txt = str_replace("'",' ',$txt) ; 
		return "<div style='overflow: hidden; height:$height; cursor: zoom-in; $style ' onClick=\"if(this.style.height=='') { this.style.height='$height'; this.style.cursor='zoom-in'; } else { this.style.height=''; this.style.cursor='zoom-out'; } \">$txt</div>" ;
}


/* Cette classe générique servant de base à des classes liées à des types de champs HTML.
	La dérivation permet principalement de bien paramétrer les différentes opérations de transformation, de validation, et de présentation en formulaire.
	Concept : chaque champ a 3 valeurs, et il faudra utiliser de manière appropriée les méthodes de transformation pour les recalculer si besoin.
	- value_sql : valeur brute, telle qu'on aimerait la lire en SQL. 
	- value_form : valeur qu'on souhaite afficher dans un formulaire, ou qu'on récupère d'un formulaire. 
	- value_html : valeur affichable dans une page d'état (ex : listing)
	Il faut noter que pour les champs à sélection (chaz_field_select_simple, chaz_field_radio_simple, chaz_field_select_keyvalue, chaz_field_radio_keyvalue)
	le choix est de facto toujours obligatoire. Il reviendra au programmeur de proposer une option "NA" (et pas une chaîne vide) si le choix est facultatif,
	et d'en faire l'option par défaut. 

*/
abstract class chaz_field {
	
	// des constantes utilisables dans les appels de fonction : chaz_field::LABEL_DISPLAY_RAW par exemple
	// elles sont publiques par essence
	
	// à utiliser avec la fonction $this->get_label() ;
	const DISPLAY_LABEL_RAW = 1 ;
	const DISPLAY_LABEL_HTML = 2 ;
	const DISPLAY_SUBTITLE_RAW = 3 ;
	const DISPLAY_SUBTITLE_HTML = 4 ;
	const DISPLAY_SQL = 6 ;
	const DISPLAY_LABEL2_RAW = 11 ;
	const DISPLAY_LABEL2_HTML = 12 ;
	
	// à utiliser avec la fonction $string=$this->transform($string) ; (valeur pas enregistrée !)
	const TRANSFORM_TRIM = 100 ;
	const TRANSFORM_NL2SPACE = 101 ;
	const TRANSFORM_NL2BR = 102 ;
	const TRANSFORM_RAW2HTML = 103 ;
	const TRANSFORM_HTML2RAW = 104 ;
	const TRANSFORM_OMITQUOTES = 105 ;
	const TRANSFORM_STR2NB = 106 ;
	const TRANSFORM_PRETTY2RAW = 107 ; //  par ex pour les SELECT jolis, avec le tableau $this->configuration['pretty2raw']
	const TRANSFORM_RAW2PRETTY = 108 ; //  par ex pour les SELECT jolis, avec le tableau $this->configuration['raw2pretty']
	const TRANSFORM_ESCAPEDOUBLEQUOTE = 109 ;
	const TRANSFORM_ESCAPESINGLEQUOTE = 110 ;
	const TRANSFORM_SUPRENDTEXTAREA = 111 ; // supprimer toute balise type </textarea
	const TRANSFORM_SIMPLIFYSPACES = 112 ; // remplace toute sles suites d'espacements par un espace
	const TRANSFORM_TONUMERIC = 113 ;
	const TRANSFORM_TOINTEGER = 114 ;
	const TRANSFORM_TOPOSITIVEINTEGER = 115 ;
	const TRANSFORM_ALLOWEDVALUESONLY = 116 ;
	const TRANSFORM_ALLBUTPROHIBITEDVALUES = 117 ;
	const TRANSFORM_DELETESPACES = 118 ; // supprime tout espace
	const TRANSFORM_SIMPLIFYPHONE = 119 ; // supprime les points, les slashes et les espaces.
	const TRANSFORM_SERIALNUMBER = 120 ; // ne garde que les chiffres et lettres, en majuscules. Idéal aussi pour IBAN, SS, etc.
	const TRANSFORM_TOLOWER = 121 ;
	const TRANSFORM_TOUPPER = 122 ;
	const TRANSFORM_DATESQL2FRENCH = 123 ;
	const TRANSFORM_DEDUPLICATE = 124 ; // pour des valeurs séparées par des virgules uniquement, dédoublonne
	const TRANSFORM_SORT = 125 ; // pour des valeurs séparées par des virgules uniquement, trie par ordre alphabétique
	const TRANSFORM_TAGIT = 126 ; // pour des valeurs séparées par des virgules uniquement, trie par ordre alphabétique
	
	// à utiliser avec la fonction $messagelist = $this->validate_value() ;
	const VALIDATE_BOOLEAN = 200 ;
	const VALIDATE_NUMERIC = 201 ;
	const VALIDATE_INTEGER = 202 ;
	const VALIDATE_POSITIVEINTEGER = 203 ;
	const VALIDATE_DATE = 204 ;
	const VALIDATE_EMAIL = 205 ;
	const VALIDATE_FRENCHZIPCODE = 206 ;
	const VALIDATE_PHONE = 207 ;
	const VALIDATE_FRENCHSOCIALPLAN = 208 ;
	const VALIDATE_IBAN = 209 ;
	const VALIDATE_LOGIN = 210 ;
	const VALIDATE_PASSWORD = 211 ;
	const VALIDATE_NOTEMPTY = 212 ;
	const VALIDATE_INALLOWEDVALUES = 213 ;
	const VALIDATE_NOTINPROHIBITEDVALUES = 214 ;
	const VALIDATE_SERIALNUMBER = 215 ; // chiffres et lettres, en majuscules
	const VALIDATE_EMAIL_UNIV_LILLE = 216 ; 
	
	
	
	// le tableau de configuration qu'on peut définir lors de la construction ou comme des champs publics
	protected $configuration = array ( 
		'title'=>"",  // Nom du champ type formulaire
		'subtitle'=>"", // Détails sur le champ, comme une consigne de remplissage
		'title2'=>"",  // Nom du champ type entête de tableau
		'default_value'=>"", // valeur par défaut, utile uniquement si création d'un nouveau champ. Toujours utilisée au début.
		'display_default_method'=>self::DISPLAY_LABEL_RAW,
		'allowed_values'=>array(),
		'prohibited_values'=>array(),
		'raw2pretty'=>array(), /* tableau associatif de transformation, type 'H'=>'Homme' */
		'pretty2raw'=>array(), /* tableau associatif de transformation, type 'Homme'=>'H' */
		'operations_form2sql'=>array(self::TRANSFORM_NL2SPACE, self::TRANSFORM_OMITQUOTES, self::TRANSFORM_TRIM, self::TRANSFORM_ALLBUTPROHIBITEDVALUES), 
								/* liste des opérations à faire automatiquement, dans cet ordre, sous forme de constantes TRANSFORM... 
								Ces opérations doivent pouvoir permettre d'enregistrer des formulaires non-valides. Tant pis, la saisie sera parfois altérée. */
		'operations_sql2form'=>array(), /* liste des opérations à faire automatiquement, dans cet ordre, sous forme de constantes TRANSFORM... */
		'operations_sql2html'=>array(self::TRANSFORM_RAW2HTML), /* liste des opérations à faire automatiquement, dans cet ordre, sous forme de constantes TRANSFORM... */
		'display_form_prefilled'=>TRUE, /* affichage d'un composant de formulaire : pré-rempli ou vierge ? */
		'display_form_error'=>FALSE, /* affichage d'un composant de formulaire : style évoquant une erreur */
		'display_form_separator'=>"<br />", /* séparateur utilisé pour les radiobox */
		'display_form_textarea_with_editor'=>FALSE, /* utiliser TinyMCE pour la textarea */
		'validations_form2sql'=>array(self::VALIDATE_NOTINPROHIBITEDVALUES), /* liste des validations à faire avant d'enregistrer une saisie de formulaire, dans cet ordre, sous forme de constantes VALIDATE... */
		'textbox_width'=>40,  /* pour affichage */
		'textarea_cols'=>70,  /* pour affichage */
		'textarea_rows'=>10,  /* pour affichage */
		
	);
	protected $value_sql ='' ;
	protected $value_form ='' ;
	protected $value_html ='' ;
	protected $error_message = '' ; // ici on stocke les messages d'erreurs enregistrés. Cela permet d'y accéder de manière asynchrone, même si ce n'est pas le mécanisme principal.
	
	
	
	/**
	 * Constructeur
	 * @param $champ
	 *        	éventuellement un nom de champ SQL
	 * @param $nom_table
	 *        	éventuellement un nom de table SQL
	 * @param array $proprietes
	 *        	éventuellement un tableau associatif de propriétés
	 */
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		if( $sql_field=='') {
			$sql_field = 'field'.rand() ;
		}
		if( $sql_table=='') {
			$sql_table = 'table'.rand() ;
		}
		$this->configuration['sql_field'] = $sql_field;
		$this->configuration['sql_table'] = $sql_table;
		foreach ( $proprietes as $clef => $valeur ) {
			$this->configuration[$clef] = $valeur;
		}
		// replir automatiquement le nom du champ, si non spécifié
		if( $this->configuration['title'] == '' ) {
			$temp = str_replace('_', ' ', $sql_field) ;
			$temp = ucwords($temp) ;
			$this->configuration['title'] = $temp ;
		}
		// ici le nom du champ pour affichage tabulaire
		if( $this->configuration['title2'] == '' ) {
			$this->configuration['title2'] = $this->configuration['title'] ;
		}
		// on met la valeur par défaut, elle sera souvent écrasée par la suite.
		$this->set_value_from_reliable_source($this->configuration['default_value']);
	}
	
	/**
	 * Destructeur
	 */
	public function __destruct() {
		//
	}
	
	/**
	 * Conversion implicite en String pour le débogage uniquement
	 *
	 * @return string Chaine HTML
	 */
	public function __toString():string {
		$chaine = "\r\n<table cellspacing=2 cellpadding=2 class=form_table>";
		foreach ( $this->configuration as $clef => $valeur ) {
			$chaine .= "<tr><td valign=top class=form_clef>$clef</td><td valign=top class=form_val>$valeur</td></tr>";
		}
		$chaine .= "</table>";
		return $chaine;
	}
	
	/**
	 * retourne le nom du champ, de diverses manières. Voir les constantes de classe.
	 *
	 * @return string Chaine HTML ou brute, selon
	 */
	public function get_label(int $way=NULL):string {
		if( is_null($way)) {
			$way = $this->configuration['display_default_method'] ;
		}
		if($way==self::DISPLAY_SQL) {
			return $nom_table.'.'.$champ ;  
		}
		if($way==self::DISPLAY_LABEL_RAW) {
			return $this->configuration['title'] ;  
		}
		if($way==self::DISPLAY_LABEL_HTML) {
			return htmlentities($this->configuration['title']) ;  
		}
		if($way==self::DISPLAY_LABEL2_RAW) {
			return $this->configuration['title2'] ;  
		}
		if($way==self::DISPLAY_LABEL2_HTML) {
			return htmlentities($this->configuration['title2']) ;  
		}
		if($way==self::DISPLAY_SUBTITLE_RAW) {
			return $this->configuration['subtitle'] ;  
		}
		if($way==self::DISPLAY_SUBTITLE_HTML) {
			return htmlentities($this->configuration['subtitle']) ;  
		}
		trigger_error("Type de retour inconnu $way", E_USER_ERROR) ; 
		return $this->configuration['title'] ; 
	}
	
	
	/* Fonction qui valide une caractéristique, et retourne une phrase si pas bon, et une chaîne vide si OK. 
	Valide une chaîne passée en paramètre, pas $this->value...
	Appel une à une. Utiliser plutôt validate_value. */
	protected function validate_feature($string, $way):string {
		// les chaînes vides sont toujours valides, sauf si on a demandé le contraire. 
		if( $way==self::VALIDATE_NOTEMPTY) {
			if( strlen(trim($string))==0 ) {
				return "Ne laissez pas cette case vide" ;
			} else {
				return "" ;
			} 
		}
		// donc passé ce point on valide toutes les chaînes vides
		if( strlen(trim($string))==0 ) {
			return "" ;
		}
		// autres modalités de validation
		if( $way==self::VALIDATE_BOOLEAN) {
			if( $string!="0" AND $string!="1"  ) {
				return "Saisissez 0 ou 1 uniquement" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_NUMERIC) {
			if( ! preg_match('/^[-]{0,1}[0-9]+\.?[0-9]*$/i', $string )) {
				return "Saisissez un nombre. Autorisés : chiffres, point, signe moins initial, sans virgule ni espace" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_INTEGER) {
			if( ! preg_match('/^[-]{0,1}[0-9]+$/i', $string )) {
				return "Saisissez un nombre entier. Autorisés : chiffres, signe moins initial, sans virgule ni point ni espace" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_POSITIVEINTEGER) {
			if( ! preg_match('/^[0-9]+$/i', $string )) {
				return "Saisissez un nombre entier. Autorisés : chiffres seulement" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_DATE) {
			if( ! preg_match('/^\d{4}-\d{2}-\d{2}$/i', $string )) {
				return "Saisissez une date au format AAAA-MM-JJ" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_EMAIL) {
			if( ! preg_match('/^(\w|-|_|\.)+@((\w|-)+\.)+([a-z]|[A-Z]){2,6}$/i', $string ) ) {
				return "Saisissez une adresse email valide, telle prenom.nom@hote.ext" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_EMAIL_UNIV_LILLE) {
			if( ! preg_match('/^(\w|-|_|\.)+@univ-lille\.fr$/i', $string ) ) {
				return "Saisissez une adresse email valide de l Université de Lille, telle prenom.nom@univ-lille.fr ou prenom.nom.etu@univ-lille.fr" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_FRENCHZIPCODE) {
			if( ! preg_match('/^[\dAB]{5}$/i', $string ) ) {
				return "Saisissez un code postal valide, tel 59000" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_PHONE) {
			if( ! preg_match('/^[\d\s\+()]{9,20}$/i', $string ) ) {
				return "Saisissez un numéro de téléphone valide. Sont autorisés : chiffres, espaces, plus, parenthèses, mais ni point ni slash ni tiret" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_FRENCHSOCIALPLAN) {
			if( ! preg_match('/^[\dA-Z]{15}$/i', $string ) ) {
				return "Saisissez un numéro de sécurité sociale valide : chiffres et parfois lettres sans espace, de longueur 15" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_IBAN) {
			if( ! preg_match('/^[A-Z]{2}[A-Z0-9]{14,32}$/i', $string ) ) {
				return "Saisissez un IBAN valide : chiffres et lettres sans espace, longueur 16 à 34" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_SERIALNUMBER) {
			if( ! preg_match('/^[\dA-Z]*$/i', $string ) ) {
				return "Saisissez un numéro de série valide : chiffres et lettres en majuscuiles sans espace" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_LOGIN) {
			if( ! preg_match('/^[A-Za-z0-9_\.@-]{6,31}$/i', $string ) ) {
				return "Saisissez un login valide : chiffres, lettres, éventuellement email, longueur de 6 à 31 " ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_PASSWORD) {
			if( ! preg_match('/^[\S]{6,31}$/i', $string ) ) {
				return "Saisissez un mot de passe valide : tout sauf espace, longueur de 6 à 31" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_INALLOWEDVALUES) {
			if( ! in_array ( $string , $this->configuration['allowed_values'] ) ) {
				$temp = implode(';', $this->configuration['allowed_values']) ;
				if( strlen($temp)>50 ) {
					$temp = substr($temp, 0, 47).'...' ;
				}
				return "Utilisez une des valeurs autorisées ($temp)" ;
			} else {
				return "" ;
			} 
		}
		if( $way==self::VALIDATE_NOTINPROHIBITEDVALUES) {
			// fonctionne aussi avec un tableau vide
			if(  in_array ( $string , $this->configuration['prohibited_values'] ) ) {
				return "N utilisez pas de valeur interdite" ;
			} else {
				return "" ;
			} 
		}
		trigger_error("Type de validation inconnu $way", E_USER_ERROR) ;
		return '' ;
	}
	
	/* Réalise la validation prévue dans la configuration, et retourne vide si OK, une chaîne de caractère si pas OK.
		Si html est true, retourne un texte HTML (à mettre entre <p> et </p> par exemple). Sinon, texte brut.
		Séparateur peut alors être spécifié : \r\n pour du texte brut, <br /> ou </li><li> pour du html par exemple.
		Il faudrait appliquer les transformations avant la validation
	*/
	public function autovalidate_value($value, $html=TRUE, $sep="<br />"): string {
		$messages = array() ;
		foreach( $this->configuration['validations_form2sql'] as $a_way ) {
			$result = $this->validate_feature($value, $a_way) ;
			if( $result != '') {
				$result = '['.$this->configuration['title'].'] '.$result ;
				if($html) {
					$messages[] = htmlentities($result) ;
				} else {
					$messages[] = $result ;
				}
				
			}
		}
		if( count($messages)==0) {
			$this->error_message = '' ;
		} else {
			$this->error_message = implode($sep, $messages) ;
		}
		return $this->error_message ;
	}
	
	
	
	/* Fonction de transformation d'une valeur à utiliser dans l'esprit d'une fonction statique */
	protected function transform($string, $way):string {
		if($way==self::TRANSFORM_TRIM) {
			return trim($string) ;
		}
		if($way==self::TRANSFORM_NL2SPACE) {
			$string = str_replace(array("\r\n", "\r", "\n"), " ", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_NL2BR) {
			$string = str_replace(array("\r\n", "\r", "\n"), "<br />", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_RAW2HTML) {
			$string = htmlentities($string) ;
			$string = str_replace(array("\r\n", "\r", "\n"), "<br />", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_HTML2RAW) {
			$string = strip_tags($string) ;
			$string = html_entity_decode($string) ;
			return $string ;
		}
		if($way==self::TRANSFORM_OMITQUOTES) {
			$string = str_replace(array("'", '"'), " ", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_STR2NB) {
			$string = trim(str_replace(" ", "", $string));
			$string = trim(str_replace(",", ".", $string));
			return (int) $string ;
		}
		if($way==self::TRANSFORM_PRETTY2RAW) {
			if( $string=='') return '' ;
			if( ! isset($this->configuration['pretty2raw'][$string]) ) {
				trigger_error("Entrée inconnue ($string) pour ". $this->configuration['sql_field'] . " dans ". $this->configuration['sql_table'], E_USER_NOTICE) ;
				return '' ;
			}
			return  $this->configuration['pretty2raw'][$string] ;
		}
		if($way==self::TRANSFORM_RAW2PRETTY) {
			if( $string=='') return 'non renseigné' ;
			if( ! isset($this->configuration['raw2pretty'][$string]) ) {
				trigger_error("Entrée inconnue ($string) pour ". $this->configuration['sql_field'] . " dans ". $this->configuration['sql_table'], E_USER_NOTICE) ;
				return '' ;
			}
			return  $this->configuration['raw2pretty'][$string] ;
		}
		if($way==self::TRANSFORM_ESCAPEDOUBLEQUOTE) {
			$string = str_replace('"', '\\"', $string);
			return $string ;
		}
		if($way==self::TRANSFORM_ESCAPESINGLEQUOTE) {
			$string = str_replace("'", "\\'", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_SUPRENDTEXTAREA) {
			$string = str_ireplace("</textarea", "<", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_SIMPLIFYSPACES) {
			$string = preg_replace("|\s+|i", " ", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_TONUMERIC) {
			$string = str_replace(",",".", $string) ;
			$string = str_replace("O","0", $string) ;
			$string = str_replace("_","", $string) ;
			// $string = preg_replace("|[^\d\-\.]|i", "", $string);
			$string = preg_replace("|\s+|i", " ", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_TOINTEGER) {
			// $string = preg_replace("|[^\d\-]|i", "", $string);
			$string = str_replace("_","", $string) ;
			$string = preg_replace("|\s+|i", " ", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_TOPOSITIVEINTEGER) {
			// $string = preg_replace("|[^\d]|i", "", $string);
			$string = str_replace("-","", $string) ;
			$string = str_replace("_","", $string) ;
			$string = preg_replace("|\s+|i", " ", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_ALLOWEDVALUESONLY) {
			if( ! in_array( $string, $this->configuration['allowed_values'] ) ) {
				return '' ;
			}
			return  $string ;
		}
		if($way==self::TRANSFORM_ALLBUTPROHIBITEDVALUES) {
			if(  in_array( $string, $this->configuration['prohibited_values'] ) ) {
				return '' ;
			}
			return  $string ;
		}
		if($way==self::TRANSFORM_DELETESPACES) {
			$string = preg_replace("|\s|i", "", $string);
			return $string ;
		}
		if($way==self::TRANSFORM_SIMPLIFYPHONE) {
			$string = preg_replace('|[\s.\x5c/-]|i', "", $string); // \x5c est le backslash. Les autres sont listés directement sans échappement. Attention, guillemets simples.
			return $string ;
		}
		if($way==self::TRANSFORM_SERIALNUMBER) {
			$string = strtoupper($string) ;
			// $string = preg_replace("|[^A-Z0-9]|i", "", $string);
			$string = preg_replace('|[\s.\x5c/-]|i', "", $string); // \x5c est le backslash. Les autres sont listés directement sans échappement. Attention, guillemets simples.
			return $string ;
		}
		if($way==self::TRANSFORM_TOUPPER) {
			$string = strtoupper($string) ;
			return $string ;
		}
		if($way==self::TRANSFORM_TOLOWER) {
			$string = strtolower($string) ;
			return $string ;
		}
		if($way==self::TRANSFORM_DATESQL2FRENCH) {
			if(strlen($string)==10) {
				$aaaa = substr($string, 0, 4) ; 
				$mm = (int) substr($string, 5, 2) ; 
				$jj = substr($string, 8, 2) ;
				if( $aaaa>0 AND $mm>0 AND $mm<=12 AND $jj>0) {
					$mois = array('jan', 'fév', 'mar', 'avr', 'mai', 'jun', 'jul', 'aoû', 'sep', 'oct', 'nov', 'déc') ;
					$mmm = $mois[($mm-1)] ;
					return $jj.' '.$mmm.' '.$aaaa ;
				}
			}
			return 'date non définie' ;
		}
		if($way==self::TRANSFORM_DEDUPLICATE) {
			$temp = explode(',',$string) ;
			$temp = array_unique($temp) ;
			$string = implode(',',$temp) ;
			return $string ;
		}
		if($way==self::TRANSFORM_SORT) {
			$temp = explode(',',$string) ;
			sort($temp) ;
			$string = implode(',',$temp) ;
			return $string ;
		}
		if($way==self::TRANSFORM_TAGIT) {
			if($string=="") {
				return $string ;
			} else {
				return "<span class=tagit_statique>". str_replace(",","</span> <span class=tagit_statique>",$string) ."</span>" ;
			}
		}
		
		
		trigger_error("Modification inconnue $way", E_USER_ERROR) ;
		return $string ;
	}
	
	/* Fonction invoquée automatiquement pour passer du formulaire au SQL. Retourne la valeur modifiée */
	public function autotransform_form2sql($value): string {
		$operations = $this->configuration['operations_form2sql'] ;
		foreach($operations as $an_operation) {
			$value = $this->transform($value, $an_operation) ;
		}
		return $value ;
	}
	
	/* Fonction invoquée automatiquement pour passer du SQL au formulaire. Retourne la valeur modifiée */
	public function autotransform_sql2form($value): string {
		$operations = $this->configuration['operations_sql2form'] ;
		foreach($operations as $an_operation) {
			$value = transform($value, $an_operation) ;
		}
		return $value ;
	}
	
	/* Fonction invoquée automatiquement pour passer du SQL à un code HTML */
	public function autotransform_sql2html($value) {
		$operations = $this->configuration['operations_sql2html'] ;
		foreach($operations as $an_operation) {
			$value = $this->transform($value, $an_operation) ;
		}
		return $value ;
	}
	
	public function get_value_sql():string {
		return $this->value_sql ;
	}
	public function get_value_html():string {
		return $this->value_html ;
	}
	public function get_value_form():string {
		return $this->value_form ;
	}
	public function get_error_message():string {
		return $this->error_message ;
	}
	
	// cette méthode permet de recalculer les valeurs form et html depuis value_sql.
	// Utile lorsque value_sql ou les paramètres de transformation ont changé entre temps
	public function update_values() {
		// je recalcule valeur_form
		$this->value_form = $this->autotransform_sql2form($this->value_sql) ;
		// je recalcule valeur_html
		$this->value_html = $this->autotransform_sql2html($this->value_sql) ;
	}
	
	// deux fonctions de SET : une d'une source fiable, l'autre d'une source FORM. Dans les deux cas : enchaîner les transformations PUIS validations => valeur SQL, puis recalculer les 2 autres valeurs.
	// cette fonction enregistre une valeur réputée fiable. Pour ce faire, elle l'enregistre en valeur SQL et recalcule les 2 autres valeurs (form et html).
	public function set_value_from_reliable_source($value_sql) {
		$value_sql = (string) $value_sql ;
		// je mets dans valeur_sql
		$this->value_sql = $value_sql ;
		// je recalcule valeur_form
		$this->value_form = $this->autotransform_sql2form($value_sql) ;
		// je recalcule valeur_html
		$this->value_html = $this->autotransform_sql2html($value_sql) ;
	}
	
	// cette fonction enregistre une valeur non-fiable, issue d'un formulaire. Elle retourne une chaîne avec toutes erreurs éventuellement relevées.
	// Elle sous-traite la suite à la fonction set_value_from_reliable_source.
	// J'ai dû "détyper" le paramètre pour être compatible avec chaz_field_file...
	public function set_value_from_unreliable_source($value_form): string {
		$value_form = (string) $value_form ; 
		// d'abord transformations automatiques
		$value_form = $this->autotransform_form2sql($value_form) ;
		// puis validation automatique
		$return_string = $this->autovalidate_value($value_form) ;
		// et si OK ou si pas OK, je conclue
		$this->set_value_from_reliable_source($value_form) ;
		// dans tous les cas je renvoie les messages de validation
		return $return_string ;
	}
	
	/**
	 * Accesseur GET pour la configuration
	 */
	public function __get($nm) {
		if (isset ( $this->configuration[$nm] )) {
			return $this->configuration[$nm];
		} else {
			return null;
		}
	}
	
	/**
	 * Accesseur SET pour la configuration
	 */
	public function __set($nm, $val):void {
		$this->configuration[$nm] = $val;
	}
	
	// fonctions de validation
	
	// fonctions d'affichage en form pré-rempli
	abstract public function get_form_component() ; 
	
	// si on veut un imput type=text
	public function get_form_component_text():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		if( ! $this->configuration['display_form_prefilled'] ) {
			return "<input type='text'  name=\"$nom\"  id=\"$nom\"  size=". $this->configuration['textbox_width'] ." $texte_css />" ;
		} else {
			return "<input type='text'  name=\"$nom\"  id=\"$nom\"  size=". $this->configuration['textbox_width'] ." value=\"$value\"  $texte_css />" ;
		}
	}
	
	// si on veut un imput type=date
	public function get_form_component_date():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		if( ! $this->configuration['display_form_prefilled'] ) {
			return "<input type='date'  name=\"$nom\"  id=\"$nom\"   $texte_css />" ;
		} else {
			return "<input type='date'  name=\"$nom\"  id=\"$nom\"  value=\"$value\"  $texte_css />" ;
		}
	}
		
	// si on veut un imput type=password
	public function get_form_component_password():string {
		$nom = $this->configuration['sql_field'] ;
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		return "<input type='password'  name=\"$nom\"  id=\"$nom\"  size=40  $texte_css />" ;
	}
	
	// si on veut un imput type=hidden
	public function get_form_component_hidden():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;
		return "<input type='hidden'  name=\"$nom\"  id=\"$nom\"  value=\"$value\" />" ;
	}
	
	// si on veut un SELECT avec un système de clefs-valeurs (traduction des libellés, tableau $this->configuration['raw2pretty'])
	public function get_form_component_select_keyvalue():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;	
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		$chaine = "\r\n<select  name=\"$nom\"  id=\"$nom\" $texte_css >" ;
		foreach( $this->configuration['raw2pretty'] as $clef=>$affichage ) {
			if( $clef == $value AND $this->configuration['display_form_prefilled'] ) {
				$sel = " selected " ;
			} else {
				$sel = " " ;
			} 
			$chaine .= "\r\n<option value=\"$clef\" $sel>$affichage</option>" ;
		}
		$chaine .= "\r\n</select>" ;
		return $chaine ;
    }	
	
	
	// si on veut un SELECT avec directement les valeurs possibles, tableau $this->configuration['allowed_values']
	public function get_form_component_select_simple():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;	
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		$chaine = "\r\n<select  name=\"$nom\"  id=\"$nom\" $texte_css >" ;
		foreach( $this->configuration['allowed_values'] as $clef ) {
			if( $clef == $value AND $this->configuration['display_form_prefilled'] ) {
				$sel = " selected " ;
			} else {
				$sel = " " ;
			} 
			$chaine .= "\r\n<option value=\"$clef\" $sel>$clef</option>" ;
		}
		$chaine .= "\r\n</select>" ;
		return $chaine ;
    }	
	
	// si on veut un RADIOBOX avec un système de clefs-valeurs (traduction des libellés, tableau $this->configuration['raw2pretty'])
	public function get_form_component_radio_keyvalue():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;	
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		$chaine = array() ;
		foreach( $this->configuration['raw2pretty'] as $clef=>$affichage ) {
			if( $clef == $value AND $this->configuration['display_form_prefilled'] ) {
				$sel = " checked " ;
			} else {
				$sel = " " ;
			}
			$chaine[]= "\r\n<input type=radio  name=\"$nom\" value=\"$clef\"  $sel $texte_css /> $affichage" ;			
		}
		$chaine = implode($this->configuration['display_form_separator'], $chaine) ;
		return $chaine ;
    }	
	
	// si on veut un RADIOBOX avec directement les valeurs possibles, tableau $this->configuration['allowed_values']
	public function get_form_component_radio_simple():string {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;	
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		$chaine = array() ;
		foreach( $this->configuration['allowed_values'] as $clef ) {
			if( $clef == $value AND $this->configuration['display_form_prefilled'] ) {
				$sel = " checked " ;
			} else {
				$sel = " " ;
			}
			$chaine[]= "\r\n<input type=radio  name=\"$nom\" value=\"$clef\"  $sel $texte_css /> $clef" ;			
		}
		$chaine = implode($this->configuration['display_form_separator'], $chaine) ;
		return $chaine ;
    }	
	
	// si on veut un TEXTAREA (penser à spécifier les méthodes de transformation)
	// avec option $this->configuration['display_form_textarea_with_editor'] pour nicEditor
	public function get_form_component_textarea():string {
		$nom = $this->configuration['sql_field'] ;
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		if( $this->configuration['display_form_prefilled'] ) {
			$valeur = $this->value_form ;	
		} else {
			$valeur='' ;
		}
		$chaine = "<textarea   name=\"$nom\"  id=\"$nom\"  cols=". $this->configuration['textarea_cols'] ." rows=". $this->configuration['textarea_rows'] ."  $texte_css>$valeur</textarea>" ;
		if( $this->configuration['display_form_textarea_with_editor'] ) {
			
			$chaine .= "\r\n<script>bkLib.onDomLoaded(function() { " ;
			$chaine .= "new nicEditor({iconsPath : './ref/nicEditorIcons.gif', buttonList : ['bold','italic','underline','left','center','right','justify','ol','ul','fontSize','fontFamily','fontFormat','indent','outdent','link','unlink','forecolor']}).panelInstance('$nom');} );" ;
			$chaine .= "\r\n</script>" ;
		} 
		return $chaine ;
    }	
	
	// si on veut un champ permettant de saisir des TAGS avec suggestion (ouverte) de valeurs possibles, 
	// via le tableau $this->configuration['allowed_values'] qui ne sera pas limitatif dans ce cas
	public function get_form_component_tags():string {
		$nom = $this->configuration['sql_field'] ;
		$valeur = $this->value_form ;	
		
		// un identifiant unique de l'objet <input> dans la page HTML
		$id_input_field = 'tag'.rand() ;
		// un nom unique de variable JavaScript qui listera les valeurs possibles
		$id_js_list = 'taglist'.rand() ;
		
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		
		$chaine = "\r\n<input name='$nom' id='$id_input_field' value='$valeur' $texte_css />" ;
		$chaine .= "\r\n<script>$(document).ready(function() { " ;
		if( count($this->configuration['allowed_values'])>0 AND $this->configuration['display_form_prefilled']) {
			$chaine .= "\r\nvar $id_js_list = ['".  implode("', '", $this->configuration['allowed_values'])  ."'];" ;
		} else {
			$chaine .= "\r\nvar $id_js_list = [];" ;
		}
		$chaine .= '$("#'.$id_input_field.'").tagit({ availableTags: '. $id_js_list .', removeConfirmation: true	});	});	</script>' ;
		return $chaine ;
    }	
	
	// Permet d'ajouter facilement (si TRUE) ou de retirer (si FALSE) l'option de validation VALIDATE_NOTEMPTY
	public function set_mandatory($way=TRUE):void {
		// pour l'ajouter, si elle n'est pas présente
		if( $way AND !in_array(self::VALIDATE_NOTEMPTY,$this->configuration['validations_form2sql']) ) {
			$this->configuration['validations_form2sql'][] = self::VALIDATE_NOTEMPTY ;
		}
		// pour la supprimer, si elle est présente (attention, array_search retourne FALSE donc position ZERO si élément absent...
		if( !$way AND in_array(self::VALIDATE_NOTEMPTY, $this->configuration['validations_form2sql']) ) {
			unset($this->configuration['validations_form2sql'][array_search(self::VALIDATE_NOTEMPTY, $this->configuration['validations_form2sql'])] );
		}
	}
	
}


class chaz_field_text_simple extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_OMITQUOTES, self::TRANSFORM_SIMPLIFYSPACES, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2HTML) ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES) ;
	}
	public function get_form_component() {
		return $this->get_form_component_text() ;
	}
}

/* on peut, optionnellement, proposer une liste d'options possibles. Pour ce faire
utiliser set_options comme d'habitude, ou une requête SQL.
Les options définies ne sont jamais restrictives : on met ce qu'on veut
dans ce composant. */
class chaz_field_tags extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_OMITQUOTES, self::TRANSFORM_SIMPLIFYSPACES, self::TRANSFORM_ALLBUTPROHIBITEDVALUES, self::TRANSFORM_DEDUPLICATE, self::TRANSFORM_SORT ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2HTML, self::TRANSFORM_TAGIT) ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES) ;
	}
	public function get_form_component() {
		return $this->get_form_component_tags() ;
	}
	// fonction qui permet de définir en une seule fois les options dans la configuration
	// $values est un tableau indexé numériquement type array('Homme', 'Femme') ;
	// Ne jamais passer d'option vide (proserait problème d'indexation du tableau)
	public function set_options(array $values) {
		$this->configuration['allowed_values'] = $values ;
	}
	public function get_options() {
		return $this->configuration['allowed_values']  ;
	}
	// la requête SQL doit retourner une seule colonne. La première colonne uniquement sera utilisée.
	public function set_options_from_sql_query(string $query, array $binding_array=array(), $bdd=NULL) {
		if( is_null($bdd)) {
			$bdd = chaz_db::getInstance() ;
		}
		$res = array() ; 
		$stt = $bdd->prepare($query) ;
		$stt->execute($binding_array) ;
		$bdd->debugue($stt) ;
		while( $tab=$stt->fetch(PDO::FETCH_NUM)) {
			$res[] = $tab[0] ; 
		}
		$this->configuration['allowed_values'] = $res ; 
	}
	/*
	class chaz_champ_tags extends chaz_champ_text {
		const SEPARATEUR = ',' ;	// mais à l'enregistrement on mettra un espace simplement pour améliorer l'aspect du mode dégradé
		protected $options = array() ;
		
		public function __construct(  $champ='', $nom_table='', array $proprietes = array()  ) {
			parent::__construct($champ, $nom_table, $proprietes) ;
			// puis nous définissons les options d'après les propriétés énoncées
			$this->bdd = new internes_db() ;
			
			if( substr( $this->configuration['type_html2'],0,4)=='SQL=' ) { // on a précisé une requête SQL
				$sql = substr( $this->configuration['type_html2'],4) ;
				$stt = $this->bdd->prepare($sql) ;
				$stt->execute() ;
				while( $ligne = $stt->fetch(PDO::FETCH_NUM) ) {
					$this->add_option($ligne[0]) ;
				}
			} else {   // on a directement précisé les valeurs possibles
				$choix = explode('|', $this->configuration['type_html2'] ) ;
				foreach( $choix as $un_choix) {
					$this->add_option($un_choix) ;
				}
			}
		}
		
		
		public function check() {
			$valeur = $_POST[$this->configuration['champ']]  ;
			$nom = $this->configuration['champ'] ;
			$libelle = $this->configuration['titre'] ;
			// vérifier si le vide est accepté
			if( $valeur === '' AND $this->configuration['obligatoire']=='1' ) {
				return "\r\n<p>Le champ $nom [$libelle] doit obligatoirement être renseigné.</p>" ;
			} else if( $valeur === '' ) {
				return '' ;
			}
			// enfin toujours valider.
			// Ici, les options proposées étant facultatives, on reformatte plutôt les saisies qu'on ne les valide.
			// cette modification sera faite dans mod_post
			return "" ;
		}
		public function mod_post() {
			$valeur = $_POST[$this->configuration['champ']] ;
			$valeur = stripslashes( $valeur ) ;
			$valeurs = explode(self::SEPARATEUR, $valeur) ;
			for($i=0; $i<count($valeurs); $i++) {
				$valeurs[$i] = trim($valeurs[$i]) ;
				$valeurs[$i] = strtolower($valeurs[$i]) ;
				$valeurs[$i] = str_replace(' ','_', $valeurs[$i]) ;
				$valeurs[$i] = str_replace('"','_', $valeurs[$i]) ;
				$valeurs[$i] = str_replace("'",'_', $valeurs[$i]) ;
				$valeurs[$i] = str_replace("]",'_', $valeurs[$i]) ;
				$valeurs[$i] = str_replace("[",'_', $valeurs[$i]) ;
			}
			sort($valeurs) ;
			$valeurs = array_unique($valeurs) ;
			return implode(self::SEPARATEUR.' ',$valeurs) ;
		}
		public function add_option($valeur) {
			$this->options[] = $valeur ;
			return true ;
		}
		public function get_affichage_valeur() {
			$valeur = $this->configuration['VALEUR'] ;
			$nom = $this->configuration['champ'] ;
			// un identifiant unique de l'objet <input> dans la page HTML
			$id_input_field = 'tag'.rand() ;
			// un nom unique de variable JavaScript qui listera les valeurs possibles
			$id_js_list = 'taglist'.rand() ;
			
			$chaine = "\r\n<input name='$nom' id='$id_input_field' value='$valeur' />" ;
			$chaine .= "\r\n<script>$(document).ready(function() { " ;
			if( count($this->options)>0) {
				$chaine .= "\r\nvar $id_js_list = ['".  implode("', '", $this->options)  ."'];" ;
			} else {
				$chaine .= "\r\nvar $id_js_list = [];" ;
			}
			$chaine .= '$("#'.$id_input_field.'").tagit({ availableTags: '. $id_js_list .', removeConfirmation: true	});	});	</script>' ;
			return $chaine ;
		}
	}
	*/
	
}



class chaz_field_hidden extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_OMITQUOTES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2HTML) ;
		$this->configuration['validations_form2sql']=array() ;
	}
	public function get_form_component() {
		return $this->get_form_component_hidden() ;
	}
}
class chaz_field_text_noneditable extends chaz_field {
	// pour un champ qui serait toujours renseigné par le programme, comme une date de dernière connexion ou ID...
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array() ;  // jamais enregistré par l'utilisateur
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2HTML) ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES) ;
	}
	public function get_form_component() {
		return $this->value_html ;
	}
}
class chaz_field_date extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_DELETESPACES, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_DATESQL2FRENCH) ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_DATE  ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=5 ; 
	}
	public function get_form_component() {
		return $this->get_form_component_date() ;
	}
}
class chaz_field_frenchzipcode extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_SERIALNUMBER, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_FRENCHZIPCODE ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=5 ; 
	}
}
class chaz_field_password extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array() ;  // surtout aucune transformation : doit être bon d'emblée, sinon recommencer la saisie.
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_PASSWORD ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=14 ; 
	}
	public function get_form_component() {
		return $this->get_form_component_password() ;
	}
}
class chaz_field_login extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_OMITQUOTES, self::TRANSFORM_DELETESPACES, self::TRANSFORM_ALLBUTPROHIBITEDVALUES, self::TRANSFORM_TOLOWER ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_LOGIN ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=14 ; 
	}
}
class chaz_field_iban extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_SERIALNUMBER, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_IBAN ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=14 ; 
	}
}
class chaz_field_frenchsocialplan extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_SERIALNUMBER, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_FRENCHSOCIALPLAN ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=14 ; 
	}
}
class chaz_field_serialnumber extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_SERIALNUMBER, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_SERIALNUMBER ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=14 ; 
	}
}
class chaz_field_phone extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_SIMPLIFYPHONE, self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_PHONE ) ;
		// pour l'affichage
		$this->configuration['textbox_width']=14 ; 
	}
}
class chaz_field_email extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_OMITQUOTES, self::TRANSFORM_SIMPLIFYSPACES, self::TRANSFORM_ALLBUTPROHIBITEDVALUES, self::TRANSFORM_TOLOWER ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES, self::VALIDATE_EMAIL ) ;
	}
}


class chaz_field_text_numeric extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['operations_form2sql'][]=self::TRANSFORM_TONUMERIC ; // en plus
		$this->configuration['validations_form2sql'][]=self::VALIDATE_NUMERIC ;
	}
}

class chaz_field_text_integer extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['operations_form2sql'][]=self::TRANSFORM_TOINTEGER ; // en plus
		$this->configuration['validations_form2sql'][]=self::VALIDATE_INTEGER  ;
	}
}

class chaz_field_text_positiveinteger extends chaz_field_text_simple {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['operations_form2sql'][]=self::TRANSFORM_TOPOSITIVEINTEGER  ; // en plus
 		$this->configuration['validations_form2sql'][]=self::VALIDATE_POSITIVEINTEGER   ;
	}
}


class chaz_field_select_simple extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ; // pas la peine de faire d'autres : fraudeur d'emblée.
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2HTML) ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_NOTINPROHIBITEDVALUES) ; // pas la peine de faire d'autres : fraudeur d'emblée.
		// pour rappel, range(1, 100) peut générer une suite d'entiers...
	}
	public function get_form_component() {
		return $this->get_form_component_select_simple() ;
	}
	// fonction qui permet de définir en une seule fois les options dans la configuration
	// $values est un tableau indexé numériquement type array('Homme', 'Femme') ;
	// On pensera à ajouter si besoin une option NA, qui ne saurait être vide (proserait problème d'indexation du tableau)
	public function set_options(array $values) {
		$this->configuration['allowed_values'] = $values ;
	}
	public function get_options() {
		return $this->configuration['allowed_values']  ;
	}
}

class chaz_field_radio_simple extends chaz_field_select_simple {
	public function get_form_component() {
		return $this->get_form_component_radio_simple() ;
	}
}

class chaz_field_select_keyvalue extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_ALLBUTPROHIBITEDVALUES ) ; // pas la peine de faire d'autres : fraudeur d'emblée. Composant propose KEY et non VALUE
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2PRETTY) ;
		$this->configuration['validations_form2sql']=array(self::VALIDATE_INALLOWEDVALUES) ; // pas la peine de faire d'autres : fraudeur d'emblée.
		// il faudra maintenir en cohérence les propriétés de configuration allowed_values et raw2pretty et pretty2raw
	}
	public function get_form_component() {
		return $this->get_form_component_select_keyvalue() ;
	}
	// fonction qui permet de définir en une seule fois les options clefs=>valeurs dans la configuration
	// $keys_values est un tableau associatif type array('H'=>'Homme', 'F'=>'Femme') ;
	// On pensera à ajouter si besoin une option NA, qui ne saurait être vide (proserait problème d'indexation du tableau)
	public function set_options(array $keys_values) {
		$this->configuration['raw2pretty'] = $keys_values ;
		$this->configuration['pretty2raw'] = array_flip($keys_values) ;
		$this->configuration['allowed_values'] = array_keys($keys_values) ;
	}
	public function get_options() {
		return $this->configuration['raw2pretty']  ;
	}
}

class chaz_field_radio_keyvalue extends chaz_field_select_keyvalue {
	public function get_form_component() {
		return $this->get_form_component_radio_keyvalue() ;
	}
}

class chaz_field_textarea_simple extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_SUPRENDTEXTAREA, self::TRANSFORM_HTML2RAW ) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array(self::TRANSFORM_RAW2HTML) ;
		$this->configuration['validations_form2sql']=array() ;
		// sans éditeur
		$this->configuration['display_form_textarea_with_editor'] = FALSE ;
	}
	public function get_form_component() {
		return $this->get_form_component_textarea() ;
	}
}

class chaz_field_textarea_editor extends chaz_field {
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		// ici quelques éléments courants de configuration
		$this->configuration['display_default_method'] = self::DISPLAY_LABEL_HTML ;
		$this->configuration['operations_form2sql']=array(self::TRANSFORM_TRIM, self::TRANSFORM_SUPRENDTEXTAREA	) ;
		$this->configuration['operations_sql2form']=array() ;  // déjà géré à l'enregistrement 
		$this->configuration['operations_sql2html']=array() ;
		$this->configuration['validations_form2sql']=array() ;
		// avec éditeur
		$this->configuration['display_form_textarea_with_editor'] = TRUE ;
	}
	public function get_form_component() {
		return $this->get_form_component_textarea() ;
	}
	
}

/* Cette classe décrit un champ de formulaire permettant l'upload d'un fichier.
Comme en vrai on stocke plusieurs informations, des fonctions supplémentaire, 
non décrites pour les autres champs,
permettent de les décrire, dont le contenu du fichier lui-même.

Notes importantes pour l'implémentation :
- les noms de variables dans le formulaire sont décrits de manière spécifique. 
  On utilisera le même nom pour une radiobox et le fichier. Laisser faire !
- pour alimenter l'entité correspondante il est fortement recommandé d'utiliser 
  le tableau chaz_field_file::merge_post_file() 
  qui parse automatiquement $_POST et $_FILES en respectant les clefs souhaitées
- le formulaire utilisé dans l'entité devra avoir un encodage "multipart/form-data"
  à imposer dans la chaz_entity correspondante
- ce type de champ ne DOIT PAS être utilisé dans les champs additionnels form:* car la 
  sérialisation de données binaires ne fonctionne pas !
- ici nous faisons une sérialisation bête avec des points virgule... 
  c'est très manuel mais ça fonctionne
 */
class chaz_field_file extends chaz_field { 
	// tableau important qui contient des infos sur le fichier
	protected $core = array('exists'=>FALSE, 'name'=>'', 'type'=>'', 'size'=>0, 'content'=>'', 'refid'=>'0') ; 
	
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		$this->configuration['max_file_size_raw'] = '5000000' ; // en octets
		$this->configuration['max_file_size_pretty'] = '5 Mo' ; // string à afficher
		$this->configuration['extensions_autorisees'] = array('pdf', 'pdfa', 'doc', 'dot', 'docx', 'dotx', 'xls', 'xlsx', 'odt', 'png', 'pneg', 'jpg', 'jpeg', 'gif' ) ; 
		
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		
		$this->configuration['operations_form2sql'] = array() ; 
		$this->configuration['operations_sql2form'] = array() ; 
		$this->configuration['operations_sql2html'] = array() ;
		$this->configuration['validations_form2sql'] = array() ; 
	}
	
	// permet de spécifier un identifiant (quelconque) qui pourra ensuite être utilisé dans get_html_code_for_download()
	// qui serait à redéfinir dans une classe fille
	public function set_refid($id) {
		$this->core['refid'] = $id ;
	}
	
	public function get_form_component() {
		$nom = $this->configuration['sql_field'] ;
		$value = $this->value_form ;
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		$texte = '<span $texte_css ><input type="hidden" name="MAX_FILE_SIZE" value="'.$this->configuration['max_file_size_raw'].'" />' ;
		if( ! $this->configuration['display_form_prefilled'] OR ! $this->file_is_defined() ) {
			$texte .= '<input type="radio" name="'.$nom.'" value="nofile" checked /> Pas de fichier<br />
				<input type="radio" name="'.$nom.'" value="create"   /> Nouveau fichier (max '.$this->configuration['max_file_size_pretty'].') : <input name="'.$nom.'" type="file"  onChange="document.forms[0].'.$nom.'.value=\'create\';"   />' ;
		} else {
			$texte .= '<input type="radio" name="'.$nom.'" value="keep" checked /> Conserver '. $this->get_html_code_for_download() .'<br />
				<input type="radio" name="'.$nom.'" value="replace"   /> Ecraser le fichier existant (max '.$this->configuration['max_file_size_pretty'].') : <input name="'.$nom.'" type="file"  onChange="document.forms[0].'.$nom.'.value=\'replace\';"   /><br />
				<input type="radio" name="'.$nom.'" value="delete" /> Supprimer le fichier existant' ;
		}
		$texte .= '</span>' ; 
		return $texte ;
	}
	
	
	
	public function get_value_html() : string {
		if( $this->file_is_defined() ) {
			$string = $this->get_html_code_for_download() ;
		} else {
			$string = '[pas de fichier]' ; 
		}
		return $string ;
	}
	
	
	// fonction qui rassemble correctement le $_POST et le $_FILE pour exploiter un formulaire
	// a condition bien entendu d'utiliser la convention de nommage des champs de formulaire...
	// la radiobox et le composant "file" doivent porter exactement le même nom. Il n'y aura pas de téléscopage.
	public static function merge_post_file() {
		$noms_fichiers = array_keys($_FILES) ;
		$resultat = array_merge($_POST, $_FILES) ;  // dans lequel l'entrée de FILES écrase celle de POST
		foreach( $noms_fichiers as $un_nom_fichier ) {
			$resultat[$un_nom_fichier]['action'] = $_POST[$un_nom_fichier] ;
		}
		return $resultat ;
	}
	
	/* Fonction qui indique si le champ comporte des erreurs, comme par exemple un fichier trop gros ou un transfert interrompu
	Tableau est de la forme : Array (
            [name] => contacts.csv
            [type] => text/csv
            [tmp_name] => C:\Users\emman\AppData\Local\Temp\php5696.tmp
            [error] => 0
            [size] => 7536
            [action] => create  )
	*/
	public function autovalidate_value($tableau, $html=TRUE, $sep="<br />"): string {
		$messages = array() ;
		/* A-t-on bien un tableau ? */
		if( ! is_array($tableau) ) {
			$messages[]='Les données du formulaire sont corrompues.' ;
		}
		/* L'action décrite est-elle valide */
		if( ! in_array($tableau['action'], array('nofile', 'create', 'keep', 'replace', 'delete') ) ) {
			$messages[]='Seule une des actions suivante peut être réalisée : nofile, create, keep, replace, delete.' ;
		}
		/* Voir les codes erreur https://www.php.net/manual/fr/features.file-upload.errors.php 		*/
		if( $tableau['action']=='create' OR $tableau['action']=='replace') {
			if( $tableau['error']==UPLOAD_ERR_INI_SIZE) {
				$messages[]='Votre fichier est trop gros (cf. configuration du serveur)' ;
			} elseif( $tableau['error']==UPLOAD_ERR_FORM_SIZE) {
				$messages[]='Votre fichier est trop gros (maximum '.$this->configuration['max_file_size_pretty'].')' ;
			} elseif( $tableau['error']==UPLOAD_ERR_PARTIAL or $tableau['error']==UPLOAD_ERR_NO_FILE) {
				$messages[]='Le transfert du fichier a échoué. Merci de recommencer.' ;
			} elseif( $tableau['error']==UPLOAD_ERR_NO_TMP_DIR or $tableau['error']==UPLOAD_ERR_CANT_WRITE or $tableau['error']==UPLOAD_ERR_EXTENSION) {
				$messages[]="L'enregistrement du fichier a échoué." ;
			}
		}
		/* Vérification de l'extension de fichier (pas le type MIME, car  ce serait seulement celui déclaré par le navigateur) */ 
		if( $tableau['action']=='create' OR $tableau['action']=='replace') {
			if( count($this->configuration['extensions_autorisees'])>0 ) {
				$extension = strtolower(pathinfo($tableau['name'], PATHINFO_EXTENSION));
				if ( ! in_array($extension, $this->configuration['extensions_autorisees'])) { 
					$temp = implode(' / ', $this->configuration['extensions_autorisees']) ;
					$messages[]="L'extension $extension n'est pas autorisée. Extensions autorisées : $temp." ;
				}
			}
		}
		// Bilan
		if( count($messages)==0) {
			$this->error_message = '' ;
		} else {
			$this->error_message = implode($sep, $messages) ;
		}
		return $this->error_message ;
	}
	
	
	public function set_value_from_reliable_source($value_sql) {
		$value_sql = (string) $value_sql ;
		if( $value_sql != '') {
			$this->core = $this->dummy_unserialize($value_sql) ;
		}
		// je mets dans valeur_sql
		$this->value_sql = $value_sql ;
		// je recalcule valeur_form
		$this->value_form = '' ; // $this->autotransform_sql2form($value_sql) ;
		// je recalcule valeur_html
		$this->value_html = '' ; // $this->autotransform_sql2html($value_sql) ;
	}
	
	// cette fonction sera à redéfinir : elle fournit le code HTML pour télécharger le fichier
	public function get_html_code_for_download(): string {
		$string = '<a>'. $this->core['name'] .' ('. pretty_file_size($this->core['size']) .')</a>' ;
		return $string ;
	}
	
	/* cette fonction enregistre une valeur non-fiable, issue d'un formulaire. 
	Elle retourne une chaîne avec toutes erreurs éventuellement relevées.
	Tableau est un ensemble de valeurs indexées
	Nous conseillons d'utiliser chaz_field_file::merge_post_file() pour obtenir 
	un MERGE du $_POST et de $_FILES bien formatté.
	Ce merge conserve parfaitement tous les autres champs du $_POST.
	Forme attendue pour tableau :  : Array (
            [name] => contacts.csv
            [type] => text/csv
            [tmp_name] => C:\Users\emman\AppData\Local\Temp\php5696.tmp
            [error] => 0
            [size] => 7536
            [action] => create  )    avec action pouvant prendre les valeurs 'nofile', 'create', 'keep', 'replace', 'delete'
	*/
	public function set_value_from_unreliable_source($tableau): string {
		// pas de transformation automatique faite avec $value_form = $this->autotransform_form2sql($value_form) ;
		// validation automatique, que le fichier soit présent ou non
		$return_string = $this->autovalidate_value($tableau) ;
		// si le champ n'est pas valide, on fait comme si le fichier n'était pas (re)défini.
		if( $return_string!='') {
			if( $this->file_is_defined() ) {
				$tableau['action']='keep' ; 
			} else {
				$tableau['action']='nofile' ; 
			}
		}
		// maintenant agissons en fonction de l'action requise
		if( $tableau['action']=='nofile' OR $tableau['action']=='delete' ) {
			$this->core = array('exists'=>FALSE, 'name'=>'', 'type'=>'', 'size'=>0, 'content'=>'', 'refid'=>0) ; 
		} elseif( $tableau['action']=='create' OR $tableau['action']=='replace' ) {
			$name = chaz_rename_file($tableau['name']) ; 
			$content = file_get_contents($tableau['tmp_name']) ;
			$size = strlen($content) ; 
			$this->core = array('exists'=>TRUE, 'name'=>$name, 'type'=>$tableau['type'], 'size'=>$size, 'content'=>$content, 'refid'=>0) ; 
		} elseif( $tableau['action']=='keep' ) {
			// action keep : devrait normalement laisser $this->core inchangé
		} else {
			// une erreur a déjà été générée précédemment par autovalidate_value
		}
		// et si OK ou si pas OK, je conclue
		$this->set_value_from_reliable_source($this->dummy_serialize($this->core)) ;
		// débogage print "<hr /><pre>" ; print_r($return_string) ; print "\n********************\n" ; print_r($tableau) ;  print "\n********************\n" ; print_r($this->core) ; print "\n********************\n" ; print_r($this->value_sql) ; print "</pre><hr />" ;
		
		// dans tous les cas je renvoie les messages de validation
		return $return_string ;
	}
	
	public function get_file_content() {
		return $this->core['content'] ;
	}
	
	public function get_file_name() {
		return $this->core['name'] ;
	}
	
	public function get_file_size() {
		return $this->core['size'] ;
	}
	
	public function file_is_defined() {
		return $this->core['exists'] ;
	}
	
	
	protected function dummy_serialize($tab) {
		$string = ($tab['exists']?'1':'0').';'.$tab['name'].';'.$tab['type'].';'.$tab['size'].';'.$tab['refid'].';'.$tab['content'] ;
		return $string ;
	}
	
	protected function dummy_unserialize($string) {
		$tab = explode(';', $string) ;
		$retour = array(
			'exists'=>($tab[0]=='1'),
			'name'=>$tab[1],
			'type'=>$tab[2],
			'size'=>$tab[3],
			'refid'=>$tab[4]
		) ; 
		// maintenant prenons tout le reste pour faire le contenu... peut contenir des ; à son tour !
		$string2 = ($retour['exists']?'1':'0').';'.$retour['name'].';'.$retour['type'].';'.$retour['size'].';'.$retour['refid'].';' ;
		$string = str_replace($string2, '', $string) ;
		$retour['content'] = $string ;
		return $retour ;
	}
}


/* Cette classe décrit un champ de formulaire permettant l'upload d'un fichier.
Comme en vrai on stocke plusieurs informations, des fonctions supplémentaire, 
non décrites pour les autres champs,
permettent de les décrire, dont le contenu du fichier lui-même.
ATTENTION : dans chaz_field_file2, le fichier doit être stocké dans une autre table.
Ce mécanisme est à gérer dans la classe d'entité.
Ici, le champ ne contiendra que des métadonnées, dont un numéro unique de fichier.

Notes importantes pour l'implémentation :
- les noms de variables dans le formulaire sont décrits de manière spécifique. 
  On utilisera le même nom pour une radiobox et le fichier. Laisser faire !
- pour alimenter l'entité correspondante il est fortement recommandé d'utiliser 
  le tableau chaz_field_file::merge_post_file() 
  qui parse automatiquement $_POST et $_FILES en respectant les clefs souhaitées
- le formulaire utilisé dans l'entité devra avoir un encodage "multipart/form-data"
  à imposer dans la chaz_entity correspondante
- ce type de champ peut être utilisé dans les champs additionnels form:*
- ici nous faisons une sérialisation bête avec des points virgule... 
  c'est très manuel mais ça fonctionne
  
Il s'agit d'un classe abstraite, qu'il faudra redéfinir pour préciser comment : 
- ajouter, mettre à jour, supprimer un fichier (dans une entité donnée)
- écrire un lien hypertexte de lecture
- il faudra aussi prévoir, dans l'entité correspondante, une méthode pour supprimer un document

 */
abstract class chaz_field_file2 extends chaz_field { 
	// tableau important qui contient des infos sur le fichier
	// id est la clef primaire, et content est le contenu binaire
	protected $core = array('exists'=>FALSE, 'name'=>'', 'type'=>'', 'size'=>0, 'id'=>'0', 'content'=>'') ; 
	
	
	public function __construct($sql_field='', $sql_table='', array $proprietes = array()) {
		// propriétés qu'on doit pouvoir modifier, mais qui seraient écrasées par des valeurs par défaut du parent
		$this->configuration['max_file_size_raw'] = '5000000' ; // en octets
		$this->configuration['max_file_size_pretty'] = '5 Mo' ; // string à afficher
		$this->configuration['extensions_autorisees'] = array('pdf', 'pdfa', 'doc', 'dot', 'docx', 'dotx', 'xls', 'xlsx', 'odt', 'png', 'pneg', 'jpg', 'jpeg', 'gif' ) ; 
		
		parent::__construct($sql_field, $sql_table, $proprietes) ;
		
		// propriétés qui passeront par-dessus les valeurs par défaut, mais qu'on ne pourra pas modifier
		$this->configuration['operations_form2sql'] = array() ; 
		$this->configuration['operations_sql2form'] = array() ; 
		$this->configuration['operations_sql2html'] = array() ;
		$this->configuration['validations_form2sql'] = array() ; 
	}
	
	
	public function get_form_component() {
		$nom = $this->configuration['sql_field'] ;
		// à cause de la présence de ':' dans certains noms de champs, 
		// on accède à la radiobox par forms[0].elements["foo"] et non forms[0].foo
		$value = $this->value_form ;
		if( $this->configuration['display_form_error'] ) {
			$texte_css = " class='error' " ;
		} else {
			$texte_css = " " ;
		}
		$texte = '<span $texte_css ><input type="hidden" name="MAX_FILE_SIZE" value="'.$this->configuration['max_file_size_raw'].'" />' ;
		if( ! $this->configuration['display_form_prefilled'] OR ! $this->file_is_defined() ) {
			
			$texte .= '<input type="radio" name="'.$nom.'" value="nofile" checked /> Pas de fichier<br />
				<input type="radio" name="'.$nom.'" value="create"   /> Nouveau fichier (max '.$this->configuration['max_file_size_pretty'].') : <input name="'.$nom.'" type="file"  onChange="document.forms[0].elements[\''.$nom.'\'].value=\'create\';"   />' ;
		} else {
			$texte .= '<input type="radio" name="'.$nom.'" value="keep" checked /> Conserver '. $this->get_html_code_for_download() .'<br />
				<input type="radio" name="'.$nom.'" value="replace"   /> Ecraser le fichier existant (max '.$this->configuration['max_file_size_pretty'].') : <input name="'.$nom.'" type="file"  onChange="document.forms[0].elements[\''.$nom.'\'].value=\'replace\';"   /><br />
				<input type="radio" name="'.$nom.'" value="delete" /> Supprimer le fichier existant' ;
		}
		$texte .= '</span>' ; 
		return $texte ;
	}
	
	
	
	public function get_value_html() : string {
		if( $this->file_is_defined() ) {
			$string = $this->get_html_code_for_download() ;
		} else {
			$string = '[pas de fichier]' ; 
		}
		return $string ;
	}
	
	
	// fonction qui rassemble correctement le $_POST et le $_FILE pour exploiter un formulaire
	// a condition bien entendu d'utiliser la convention de nommage des champs de formulaire...
	// la radiobox et le composant "file" doivent porter exactement le même nom. Il n'y aura pas de téléscopage.
	public static function merge_post_file() {
		$noms_fichiers = array_keys($_FILES) ;
		$resultat = array_merge($_POST, $_FILES) ;  // dans lequel l'entrée de FILES écrase celle de POST
		foreach( $noms_fichiers as $un_nom_fichier ) {
			$resultat[$un_nom_fichier]['action'] = $_POST[$un_nom_fichier] ;
		}
		return $resultat ;
	}
	
	/* Fonction qui indique si le champ comporte des erreurs, comme par exemple un fichier trop gros ou un transfert interrompu
	Tableau est de la forme : Array (
            [name] => contacts.csv
            [type] => text/csv
            [tmp_name] => C:\Users\emman\AppData\Local\Temp\php5696.tmp
            [error] => 0
            [size] => 7536
            [action] => create  )
	*/
	public function autovalidate_value($tableau, $html=TRUE, $sep="<br />"): string {
		$messages = array() ;
		/* A-t-on bien un tableau ? */
		if( ! is_array($tableau) ) {
			$messages[]='Les données du formulaire sont corrompues.' ;
		}
		/* L'action décrite est-elle valide */
		if( ! in_array($tableau['action'], array('nofile', 'create', 'keep', 'replace', 'delete') ) ) {
			$messages[]='Seule une des actions suivante peut être réalisée : nofile, create, keep, replace, delete.' ;
		}
		/* Voir les codes erreur https://www.php.net/manual/fr/features.file-upload.errors.php 		*/
		if( $tableau['action']=='create' OR $tableau['action']=='replace') {
			if( $tableau['error']==UPLOAD_ERR_INI_SIZE) {
				$messages[]='Votre fichier est trop gros (cf. configuration du serveur)' ;
			} elseif( $tableau['error']==UPLOAD_ERR_FORM_SIZE) {
				$messages[]='Votre fichier est trop gros (maximum '.$this->configuration['max_file_size_pretty'].')' ;
			} elseif( $tableau['error']==UPLOAD_ERR_PARTIAL or $tableau['error']==UPLOAD_ERR_NO_FILE) {
				$messages[]='Le transfert du fichier a échoué. Merci de recommencer.' ;
			} elseif( $tableau['error']==UPLOAD_ERR_NO_TMP_DIR or $tableau['error']==UPLOAD_ERR_CANT_WRITE or $tableau['error']==UPLOAD_ERR_EXTENSION) {
				$messages[]="L'enregistrement du fichier a échoué." ;
			}
		}
		/* Vérification de l'extension de fichier (pas le type MIME, car  ce serait seulement celui déclaré par le navigateur) */ 
		if( $tableau['action']=='create' OR $tableau['action']=='replace') {
			if( count($this->configuration['extensions_autorisees'])>0 ) {
				$extension = strtolower(pathinfo($tableau['name'], PATHINFO_EXTENSION));
				if ( ! in_array($extension, $this->configuration['extensions_autorisees'])) { 
					$temp = implode(' / ', $this->configuration['extensions_autorisees']) ;
					$messages[]="L'extension $extension n'est pas autorisée. Extensions autorisées : $temp." ;
				}
			}
		}
		// Bilan
		if( count($messages)==0) {
			$this->error_message = '' ;
		} else {
			$this->error_message = implode($sep, $messages) ;
		}
		return $this->error_message ;
	}
	
	/* Cette méthode ne se préoccupe que de métadonnées, et non du contenu du fichier.
		Pour récupérer le contenu du fichier, utiliser retrieve_file_content() puis get_file_content()
	*/
	public function set_value_from_reliable_source($value_sql) {
		
		$value_sql = (string) $value_sql ;
		if( $value_sql != '') {
			$this->core = $this->dummy_unserialize($value_sql) ;
		}
		// je mets dans valeur_sql
		$this->value_sql = $value_sql ;
		// je recalcule valeur_form
		$this->value_form = '' ; // $this->autotransform_sql2form($value_sql) ;
		// je recalcule valeur_html
		$this->value_html = '' ; // $this->autotransform_sql2html($value_sql) ;
	}
	
	/* cette fonction sera à redéfinir : elle fournit le code HTML pour télécharger le fichier
	   exemple d'implémentation : 
	   return '<a>'. $this->core['name'] .' ('. pretty_file_size($this->core['size']) .')</a>' ;
	*/
	abstract public function get_html_code_for_download(): string ; 
	
	
	/* cette fonction enregistre une valeur non-fiable, issue d'un formulaire. 
	Elle retourne une chaîne avec toutes erreurs éventuellement relevées.
	Tableau est un ensemble de valeurs indexées
	Nous conseillons d'utiliser chaz_field_file::merge_post_file() pour obtenir 
	un MERGE du $_POST et de $_FILES bien formatté.
	Ce merge conserve parfaitement tous les autres champs du $_POST.
	Forme attendue pour tableau :  : Array (
            [name] => contacts.csv
            [type] => text/csv
            [tmp_name] => C:\Users\emman\AppData\Local\Temp\php5696.tmp
            [error] => 0
            [size] => 7536
            [action] => create  )    avec action pouvant prendre les valeurs 'nofile', 'create', 'keep', 'replace', 'delete'
	
	Cette fonction enregistre également le contenu du fichier ainsi chargé, voir fin du script. 
	*/
	public function set_value_from_unreliable_source($tableau): string {
		// pas de transformation automatique faite avec $value_form = $this->autotransform_form2sql($value_form) ;
		// validation automatique, que le fichier soit présent ou non
		$return_string = $this->autovalidate_value($tableau) ;
		// si le champ n'est pas valide, on fait comme si le fichier n'était pas (re)défini.
		if( $return_string!='') {
			if( $this->file_is_defined() ) {
				$tableau['action']='keep' ; 
			} else {
				$tableau['action']='nofile' ; 
			}
		}
		// maintenant agissons en fonction de l'action requise
		if( $tableau['action']=='nofile' OR $tableau['action']=='delete' ) {
			$old_id = $this->core['id'] ;
			$this->core = array('exists'=>FALSE, 'name'=>'', 'type'=>'', 'size'=>0, 'content'=>'', 'id'=>0) ; 
		} elseif( $tableau['action']=='create'  ) {
			$name = chaz_rename_file($tableau['name']) ; 
			$content = file_get_contents($tableau['tmp_name']) ;
			$size = strlen($content) ; 
			$id = self::generate_serial(50) ;
			$this->core = array('exists'=>TRUE, 'name'=>$name, 'type'=>$tableau['type'], 'size'=>$size, 'content'=>$content, 'id'=>$id) ; 
		} elseif( $tableau['action']=='replace' ) {
			$name = chaz_rename_file($tableau['name']) ; 
			$content = file_get_contents($tableau['tmp_name']) ;
			$size = strlen($content) ; 
			$this->core = array('exists'=>TRUE, 'name'=>$name, 'type'=>$tableau['type'], 'size'=>$size, 'content'=>$content, 'id'=>$this->core['id']) ; 
		} elseif( $tableau['action']=='keep' ) {
			// action keep : devrait normalement laisser $this->core inchangé
		} else {
			// une erreur a déjà été générée précédemment par autovalidate_value
		}
		// et si OK ou si pas OK, je conclue
		// print "<hr /><pre>" ; print_r($return_string) ; print "\n********************\n" ; print $tableau['action'] ; print "\n********************\n" ;  print_r($tableau) ;  print "\n********************\n" ; print_r($this->core) ; print "\n********************\n" ; print_r($this->value_sql) ; print "</pre><hr />" ;
		
		// maintenant on enregistre le contenu du fichier, à l'aide des méthodes redéfinies
		if( $tableau['action']=='delete' ) {
			// renseigner provisoirement l'ID, pour la délétion
			$this->core['id'] = $old_id ;
			$return_string .= $this->delete_file_content() ;
			$this->core['id'] = 0 ;
		} elseif( $tableau['action']=='create' OR $tableau['action']=='replace' ) {
			$return_string .= $this->register_file_content() ;
		}
		
		// ENSUITE seulement on enregistre le champ : à cette occasion, $this->core['content'] est perdu
		$this->set_value_from_reliable_source($this->dummy_serialize($this->core)) ;
		
		// dans tous les cas je renvoie les messages de validation
		return $return_string ;
	}
	
	
	public static function generate_serial($length=255) {
		$caracteres = array_merge(range('a', 'z'), range('A', 'Z'), range(0, 9)) ;
		$str = '' ; 
		for( $i=0; $i<$length; $i++) {
			$offset = rand( 0 , count($caracteres)-1 ) ;
			$str .= $caracteres[$offset] ;
		}
		return $str ;
	}
	
	public function get_file_content() {
		return $this->core['content'] ;
	}
	
	public function get_file_id() {
		return $this->core['id'] ;
	}
	
	public function get_file_name() {
		return $this->core['name'] ;
	}
	
	public function get_file_size() {
		return $this->core['size'] ;
	}
	
	public function get_file_size_pretty() {
		return pretty_file_size($this->core['size']) ;
	}
	
	public function file_is_defined() {
		return $this->core['exists'] ;
	}
	
	
	protected function dummy_serialize($tab) {
		$string = ($tab['exists']?'1':'0').';'.$tab['name'].';'.$tab['type'].';'.$tab['size'].';'.$tab['id'] ;
		return $string ;
	}
	
	protected function dummy_unserialize($string) {
		$tab = explode(';', $string) ;
		$retour = array(
			'exists'=>($tab[0]=='1'),
			'name'=>$tab[1],
			'type'=>$tab[2],
			'size'=>$tab[3],
			'id'=>$tab[4]
		) ; 
		return $retour ;
	}
	
	/* méthode qui enregistre le contenu du fichier comme on souhaite : fichier physique, BDD, etc.
	   à noter que les métadonnées sont déjà gérées par ailleurs.
	   on conseille une méthode du type "replace", qui peut indifféremment créer ou écraser un enregistrement
	   le contenu est dans $this->core['content'] ;
	   Doit retourner un message d'erreur en string, ou une chaîne vide.
	*/
	abstract public function register_file_content($options=array()) : string ;
	
	/* méthode qui récupère le contenu du fichier comme on souhaite : fichier physique, BDD, etc.
	   à noter que les métadonnées sont déjà gérées par ailleurs.
	   le contenu est placé dans $this->core['content'] mais pas retourné
	   Il faudra ensuite utiliser get_file_content pour le récupérer dans une variable externe
	*/
	abstract public function retrieve_file_content($options=array()) ;
	
	/* méthode qui supprime le contenu du fichier comme on souhaite : fichier physique, BDD, etc.
	   à noter que les métadonnées sont déjà gérées par ailleurs.
	   Doit retourner un message d'erreur en string, ou une chaîne vide.
	*/
	abstract public function delete_file_content($options=array()) : string ;
	
}

/*
Cette classe permet de décrire une table SQL entière : 
c'est donc un repository d'objets (lignes).
Pas de jointure en soi. Cependant, les FK restent accessibles dans les entités, et on programmera ce qu'on veut.
Cette classe n'a pas à connaître les champs disponibles dans les entités.

Chaque repository est un SINGLETON : il faudra dériver une classe par table.
Les choses à dériver en particulier sont : 
- le nom de la table en question => redéfinir la fonction __construct qui est en PRIVATE !!!
- les méthodes créant des objets entités : classe différente

*/

abstract class chaz_entity_repository {
	private static $_instance = null;
	protected $tablename ;
	protected $pkname ;
	protected $db ; // SINGLETON de classe chaz_db (extends PDO) 
	protected $classname = "" ; // le nom de la classe d'entité (pas du repository), qui sera utilisé directement pour certaines instanciations.
	protected $tracetable_for_delete = "" ;  // éventuellement, une table SQL de même format que la table d'intérêt, dans laquelle on souhaite enregistrer les effacements de lignes
	
	private function __construct($tablename, $pkname="id") {
		$this->tablename = $tablename ;
		$this->pkname = $pkname ;
		// objet singleton à garder sous la main
		$this->db = chaz_db::getInstance() ;
	}
	
	
	// à redéfinir : méthode qui crée l'unique instance de la classe si elle n'existe pas encore puis la retourne. Pattern SINGLETON.
	abstract public static function getInstance() ;
		
	public function db_drop_one_with_id($pk): bool {
		if( $this->tracetable_for_delete != "" ) {
			$query = "INSERT INTO ". $this->tracetable_for_delete ." SELECT * FROM ". $this->tablename ." WHERE ". $this->pkname ." = :id ;" ;
			$statement = $this->db->prepare($query) ;
			$statement->execute(array( ':id' => $pk )) ; 
		}
		$query = "DELETE FROM ". $this->tablename ." WHERE ". $this->pkname ." = :id ;" ;
		$statement = $this->db->prepare($query) ;
		$statement->execute(array( ':id' => $pk )) ; 
		$res = (boolean) $statement->rowCount();
		return $res ; 
	}
	
	public function db_select_one_item_with_pk($pk) : ? chaz_entity {
		$classname = $this->classname ; 
		$query = "select * from ".$this->tablename." where ".$this->pkname." = :id ;" ;
		$stt = $this->db->prepare($query) ;
		$stt->execute(array(':id'=>$pk)) ;
		if( $tab=$stt->fetch(PDO::FETCH_ASSOC)) {
			$obj = new $classname($pk) ;
			$obj->fill_from_reliable_array($tab) ; 
			return $obj ; 
		} else {
			trigger_error("Objet de classe= $classname et PK= $pk introuvable.", E_USER_NOTICE) ; 
			return NULL ; 
		}
	}
	
	public function db_select_one_item_with_query($query, $binding_array) : ? chaz_entity {
		$classname = $this->classname ; 
		$stt = $this->db->prepare($query) ;
		$stt->execute($binding_array) ;
		if( $tab=$stt->fetch(PDO::FETCH_ASSOC)) {
			$pk = $tab[$this->pkname] ;
			$obj = new $classname($pk) ;
			$obj->fill_from_reliable_array($tab) ; 
			return $obj ; 
		} else {
			// trigger_error("aucun $classname retourné par $query avec ".serialize($binding_array), E_USER_NOTICE) ; 
			return NULL ; 
		}
	}
	
	public function db_select_many_items_with_query($query, $binding_array) : array {
		$classname = $this->classname ; 
		$res = array() ; 
		$stt = $this->db->prepare($query) ;
		$stt->execute($binding_array) ;
		$this->db->debugue($stt) ;
		while( $tab=$stt->fetch(PDO::FETCH_ASSOC)) {
			$pk = $tab[$this->pkname] ;
			$obj = new $classname($pk) ;
			$obj->fill_from_reliable_array($tab) ; 
			$res[] = $obj ; 
		}
		return $res ; 
	}
	
	// la même que db_select_many_items_with_query mais résultat indexé par ID
	public function db_select_many_items_with_query_indexed($query, $binding_array) {
		$classname = $this->classname ; 
		$res = array() ; 
		$stt = $this->db->prepare($query) ;
		$stt->execute($binding_array) ;
		$this->db->debugue($stt) ;
		while( $tab=$stt->fetch(PDO::FETCH_ASSOC)) {
			$pk = $tab[$this->pkname] ;
			$obj = new $classname($pk) ;
			$obj->fill_from_reliable_array($tab) ; 
			$res[$pk] = $obj ; 
		}
		return $res ; 
	}
	
	public function db_select_all_items(): array {
		$query = "SELECT * FROM   ". $this->tablename ." ; " ;
		return $this->db_select_many_items_with_query($query, array()) ;
	}
	
	public function get_tablename() {
		return $this->tablename ;
	}
	
	public function get_pkname() {
		return $this->pkname ;
	}
	
}


/*
Cette classe sera instanciée pour chaque objet : ligne de tableau, ou nouvel objet vierge.
*/

abstract class chaz_entity {
	protected $repository ; // le SINGLETON de la classe de repository correspondante.
	protected $db ; // le SINGLETON de classe chaz_db (extends PDO) 
	protected $pk_value ; // valeur de PK de cette instance précise (cette ligne de la table). ou NULL en cas de création pré-BDD
	protected $fields = array() ; // un tableau de champs de classes chaz_field... ils devraient être utilisés pour les GET et les SET, à redéfinir un à un. Objets indexés par le nom de champ SQL.
	protected $parite_ligne_form = 1 ; // un indicateur binaire pour savoir si on est sur une ligne paire ou impaire, donc si on change de couleur ou non.
	protected $form_encoding = 'application/x-www-form-urlencoded' ; // penser à mettre "multipart/form-data" si upload de fichiers
	
	// pas besoin de la dériver
	public function __construct($pk_value=NULL) {
		$this->pk_value = $pk_value ;
		// objets singleton à garder sous la main
		$this->db = chaz_db::getInstance() ;
		// compléments de construction, à dériver
		$this->complements_construct() ;
	}
	
	
	public function get_id() {
		return $this->pk_value ;
	}
	
	// pour présenter sous forme de liste de champs non-personnalisée
	// il est plutôt conseillé de le faire à la main dans les pages concernées
	// $complet indique si la méthode retourne aussi les balises d'encadrement <table>
	// limit_fields permet, si on le souhaite, de limiter les champs affichés.
	public function display_html_tab($complet=TRUE, $limit_fields=array(), $avec_soustitres=FALSE ) {
		$this->parite_ligne_form = 1 ;
		$str = "" ;
		if($complet) {
			$str = "<table border=0>" ;
		}
		if( count($limit_fields) > 0 ) {
			$keys = array_intersect( $limit_fields, array_keys($this->fields)) ;
		} else {
			$keys = array_keys($this->fields) ;
		}
		
		foreach( $keys as $fieldname ) {
			$str .= $this->display_html_tab_field($fieldname, $avec_soustitres) ;
		}
		if($complet) {
			$str .= "</table>" ;
		}
		return $str ; 
	}
	
	// pour présenter sous forme de liste de champs non-personnalisée, ici seulement une ligne correspondant à un champ.
	public function display_html_tab_field($fieldname, $avec_soustitres=TRUE): string {
		$str = "" ;
		$obj = $this->get_field($fieldname) ;
		if( is_null($obj))  {
			return "" ;
		} else {
			$this->parite_ligne_form = 1 - $this->parite_ligne_form ;
			$class = "form_ligne". $this->parite_ligne_form ;
			$str .= "<tr><td width=25% class='$class'><p style=text-align:left;>". $obj->get_label( chaz_field::DISPLAY_LABEL_HTML ) ;
			if($avec_soustitres) {
				$str .= "<br /><span style=font-size:0.7em >" . $obj->get_label( chaz_field::DISPLAY_SUBTITLE_RAW ) ."</span>" ;
			}
			$str .= "</p></td><td class='$class'>". $obj->get_value_html() ."</td></tr>" ;
			return $str ; 
		}
	}
	
	
	// pour présenter sous forme de formulaire pré-rempli
	// $complet indique si la sortie contient déjà les balises <form> et <table> et le bouton de validation.
	// il est plutôt conseillé de le faire à la main dans les pages concernées, notamment pour améliorer la disposition
	// limit_fields permet éventuellement de nommer les champs qu'on souhaite afficher (restriction), ou de facto d'imposer leur ordre.
	public function display_html_form($complet=TRUE, $display_errors=TRUE, $limit_fields=array() ) {
		$this->parite_ligne_form = 1 ;
		if( count($limit_fields) > 0 ) {
			$keys = array_intersect( $limit_fields, array_keys($this->fields)) ;
		} else {
			$keys = array_keys($this->fields) ;
		}
		$str = "" ;
		if( $complet ) {
			$str = "<form method='POST' enctype='".$this->form_encoding."' ><table border=0>" ;
		}
		foreach( $keys as $fieldname ) {
			$str .= $this->display_html_form_field($fieldname, $display_errors ) ;
		}
		if( $complet ) {
			$str .= "</table><p>(*) Les champs marqués d'un astérisque sont obligatoires.</p><input type='submit' value='Enregistrer' /></form>" ;
		}
		return $str ; 
	}
	
	// retourne une ligne de tableau permettant d'afficher un formulaire pour le champ demandé.
	public function display_html_form_field($fieldname, $display_errors=TRUE ): string {
		$obj = $this->get_field($fieldname) ;
		if( is_null( $obj )) {
			return "" ;
		} else {
			if( get_class ($obj) != "chaz_field_hidden") {
				$this->parite_ligne_form = 1 - $this->parite_ligne_form ;
				$class = "form_ligne". $this->parite_ligne_form ;
				$str = "<tr><td valign=top width=25% class='$class'><p style=text-align:left;>". $obj->get_label( chaz_field::DISPLAY_LABEL_HTML ) ."<br /><span style=font-size:0.7em >"
					. $obj->get_label( chaz_field::DISPLAY_SUBTITLE_RAW ) ."</span></p></td><td valign=top class='$class'>". $obj->get_form_component()  ;
				if( $obj->get_error_message() !='' ) {
					$str .= "<div class='alert_danger'>". $obj->get_error_message() ."</div>" ;
				}
				$str ."</td></tr>" ;
			} else {
				$str .= $obj->get_form_component()  ;
			}
		}
		return $str ; 
	}
	
	
	
	public function new_values_into_db( $limit_fields=array() ) {
		$tablename = $this->repository->get_tablename() ;
		$pkname = $this->repository->get_pkname() ;
		if( count($limit_fields) > 0 ) {
			$fieldnames = array_intersect( $limit_fields, array_keys($this->fields)) ;
		} else {
			$fieldnames = array_keys($this->fields) ;
		}
		$fieldnames = array_diff($fieldnames, array($pkname) ) ;
		$query = "INSERT INTO $tablename ( `". implode('` , `',$fieldnames) . "` ) VALUES ( " ; 
		$temp = array() ;
		$binding_array = array() ;
		foreach($fieldnames as $a_field_name) {
			$a_field_name2 = str_replace(':', '___', $a_field_name) ;
			$temp[] = " :$a_field_name2 " ;
			$binding_array[":".$a_field_name2] = $this->fields[$a_field_name]->get_value_sql() ; 
		}
		$query .= implode(',', $temp )  ;
		$query .= " ) ; " ;
		$stt = $this->db->prepare($query) ;
		$stt->execute($binding_array) ;
		$this->db->debugue($stt) ;
		// print "<hr /><pre>$query\n\n" ; print_r($limit_fields) ; print_r( $binding_array ) ; print_r( $_POST ) ; print "</pre><hr />" ;
		return TRUE ;
	}
	
	public function save_values_to_db( $limit_fields=array() ) {
		$tablename = $this->repository->get_tablename() ;
		$pkname = $this->repository->get_pkname() ;
		if( count($limit_fields) > 0 ) {
			$fieldnames = array_intersect( $limit_fields, array_keys($this->fields)) ;
		} else {
			$fieldnames = array_keys($this->fields) ;
		}
		$fieldnames = array_diff($fieldnames, array($pkname) ) ;
		$query = "UPDATE $tablename SET " ; 
		$temp = array() ;
		$binding_array = array() ;
		foreach($fieldnames as $a_field_name) {
			$a_field_name2 = str_replace(':', '___', $a_field_name) ;
			$temp[] = " `$a_field_name` = :$a_field_name2 " ;
			$binding_array[":".$a_field_name2] = $this->fields[$a_field_name]->get_value_sql() ; 
		}
		$query .= implode(',', $temp )  ;
		$query .= " WHERE $pkname =  ". $this->pk_value . " ; " ; // pas joli de faire en dur, mais cela évite tout téléscopage de BIND avec les champs autres...
		$stt = $this->db->prepare($query) ;
		$stt->execute($binding_array) ;
		$this->db->debugue($stt ) ;
		// print "<hr /><pre>$query\n\n" ; print_r( $binding_array ) ; print_r( $_POST ) ; print "</pre><hr />" ;  
		return TRUE ;
	}
	
	// utilisation des étiquettes (statiques) des champs, en accès par leur nom. Pour générer une jolie page.
	
	// le titre du champ en HTML, pour un affichage vertical
	public function get_field_title(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname] ) ) {
			return $this->fields[$fieldname]->get_label(chaz_field::DISPLAY_LABEL_HTML) ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	// le sous-titre du champ en HTML, pour un affichage vertical
	public function get_field_subtitle(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname] ) ) {
			return $this->fields[$fieldname]->get_label(chaz_field::DISPLAY_SUBTITLE_HTML) ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	// le titre du champ en HTML, pour un affichage tabulaire horizontal (généralement plus court)
	public function get_field_title2(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname] )) {
			return $this->fields[$fieldname]->get_label(chaz_field::DISPLAY_LABEL2_HTML) ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	
	
	// manipulation des valeurs des champs, en accès par leur nom. Pour voir le détail de ces fonctions, voir la classe chaz_field.
	// Dans la dérivation, on peut en profiter pour "intercepter" certaines de ces commandes et les modifier, si on veut.
	// Mais il vaut mieux définir des méthodes GET spécifiques et explicites plutôt, dans l'entité directement.
	
	// obtenir la valeur brute, telle qu'enregistrable en BDD
	public function get_field_value_sql(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname]) ) {
			return $this->fields[$fieldname]->get_value_sql() ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	// obtenir la valeur HTML, pour un listing par exemple
	public function get_field_value_html(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname]) ) {
			return $this->fields[$fieldname]->get_value_html() ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	// autre manière d'obtenir la valeur HTML ou SQL d'un champ
	// Par défaut c'est la valeur HTML. Un underscore final oriente vers la valeur SQL.
	// Si le champ n'est pas trouvé, on tente de remplacer le premier "form_" par "form:"
	public function __get($fieldname) {
		$fieldname1 = $fieldname ;
		if( substr($fieldname,-1)=='_' ) {
			$fieldname2 = substr($fieldname,0, -1) ;
		} else {
			$fieldname2 = FALSE ;
		}
		if( substr($fieldname,0,5)=='form_' ) {
			$fieldname3 = 'form:' . substr($fieldname,5) ;
		} else {
			$fieldname3 = FALSE ;
		}
		if( $fieldname3 AND substr($fieldname3,-1)=='_' ) {
			$fieldname4= substr($fieldname3,0, -1) ;
		} else {
			$fieldname4 = FALSE ;
		}
		
		if( $fieldname1 AND isset( $this->fields[$fieldname1]) ) {
			return $this->fields[$fieldname1]->get_value_html() ;
		} 
		if( $fieldname2 AND isset( $this->fields[$fieldname2]) ) {
			return $this->fields[$fieldname2]->get_value_sql() ;
		} 
		if( $fieldname3 AND isset( $this->fields[$fieldname3]) ) {
			return $this->fields[$fieldname3]->get_value_html() ;
		} 
		if( $fieldname4 AND isset( $this->fields[$fieldname4]) ) {
			return $this->fields[$fieldname4]->get_value_sql() ;
		} 
		
		else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	
	// obtenir la valeur affichable dans un formulaire. ATTENTION, on préfèrera bien entendu obtenir directement le code HTML du COMPOSANT FORMULAIRE, cf. post.
	public function get_field_value_form(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname] )) {
			return $this->fields[$fieldname]->get_value_form() ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	// obtenir le composant de formulaire déjà rempli
	public function get_field_form_component(string $fieldname): ? string {
		if( isset( $this->fields[$fieldname] )) {
			return $this->fields[$fieldname]->get_form_component() ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	
	// obtenir l'objet champ, pour je ne sais quel besoin...
	public function get_field(string $fieldname): ? chaz_field {
		if( isset( $this->fields[$fieldname] )) {
			return $this->fields[$fieldname] ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	
	// enregistrer une valeur fiable (ex : issue d'une BDD)
	public function set_field_value_from_reliable_source(string $fieldname, string $value_sql) {
		if( isset( $this->fields[$fieldname] )) {
			return $this->fields[$fieldname]->set_value_from_reliable_source($value_sql) ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	// enregistrer une valeur non fiable, par exemple issue d'un formulaire
	public function set_field_value_from_unreliable_source(string $fieldname, string $value_form): string {
		if( isset( $this->fields[$fieldname] )) {
			return $this->fields[$fieldname]->set_value_from_unreliable_source($value_form) ;
		} else {
			$trace = debug_backtrace();
			trigger_error('Champ non-défini : '.$fieldname.' dans '. $trace[0]['file'].' ligne '.$trace[0]['line'],E_USER_NOTICE);
			return NULL ;
		}
	}
	
	
	
	// dans cette méthode, à dériver obligatoirement, on mettra qqch comme :
	// 	$this->repository = myrepository::getInstance() ; (pattern singleton)
	// et la déclaration et le référencement des champs, dans $this->fields
	abstract public function complements_construct() ;
	
	
	// pour le remplissage depuis un select * par exemple
	// limit_fields permet, si on le souhaite, de restreindre l'analyse de $assoc_array à certains champs
	public function fill_from_reliable_array(array $assoc_array, array $limit_fields=array() )  {
		$keys = array_keys($assoc_array) ; 
		$keys = array_intersect($keys, array_keys($this->fields)) ;
		if( count($limit_fields)>0 ) {
			$keys = array_intersect($keys, $limit_fields) ;
		}
		foreach($keys as $key) {
			$this->fields[$key]->set_value_from_reliable_source($assoc_array[$key]) ;
		}
	}
	
	// pour le remplissage depuis un POST par exemple
	// limit_fields permet, si on le souhaite, de restreindre l'analyse de $assoc_array à certains champs
	// Retourne un tableau de STRINGS avec les erreurs relevées. 
	public function fill_from_unreliable_array(array $assoc_array, array $limit_fields=array() ): array  {
		$messages = array() ; 
		$keys = array_keys($assoc_array) ; 
		$keys = array_intersect($keys, array_keys($this->fields)) ;
		if( count($limit_fields)>0 ) {
			$keys = array_intersect($keys, $limit_fields) ;
		}
		foreach($keys as $key) {
			$un_message = $this->fields[$key]->set_value_from_unreliable_source($assoc_array[$key]) ;
			if( !is_null($un_message)  AND $un_message != "" ) {
				$messages[] = $un_message ;
			}
		}
		return $messages ; 
	}
	
	

	
	// peut être redéfinie, de manière facultative, si on veut faire des DELETE CASCADE...
	// (si on veut systématiquement tracer une délétion dans un fichier de Log, pas besoin de redéfinir, il suffit de renseigner $lerepository->tracetable_for_delete 
	public function delete_in_db(): bool {
		return $this->repository->db_drop_one_with_id($this->pk_value) ; 
	}
	
	
		
	
	
}


/*
Dans cette classe, il faut qu'un champ SQL (non-déclaré côté PHP) s'appelle form, 
et que plusieurs champs PHP s'appellent "form:*".
Alors, tous les champs PHP "form:*" seront sérialisés dans le champ SQL "form".
Inversement, le champ SQL "form" sera désérialisé en ensemble de champs PHP "form:*".
A noter l'usage du mot clef static:: (qui cherche dans la classe fille) au lieu de self:: (qui reste dans la classe mère)
Ce problème de portée (de l'invocation de get_additional_fieldnames() ) est la raison pour laquelle, autrefois, je recopiais ce code.
*/


abstract class chaz_entity_form extends chaz_entity {
	
	// à redéfinir pour lister les champs à sérialiser dans "form"
	// il faudra l'appeler avec "static::" et surtout pas avec "self::"
	public static function get_additional_fieldnames(): array {
		return array() ;
	}
		
	// Ici il s'agit de PHP -> BDD, alors il faut sérialiser "form:*" vers "form"
	// il faudrait toujours définir limit_fields, car actuellement la sérialisation ne modifie pas ce tableau
	public function save_values_to_db( $limit_fields=array() ) {
		$final_tab = array() ;
		$field_list = static::get_additional_fieldnames() ;
		foreach( $field_list as $a_fieldname) {
			$final_tab[$a_fieldname] = $this->fields[$a_fieldname]->get_value_sql() ;
		}
		$this->fields['form']->set_value_from_reliable_source( serialize( $final_tab ) ) ;
		if( count($limit_fields)>0) {
			// alors on retire les champs à sérialiser et on ajoute "form"
			$limit_fields = array_diff( $limit_fields, array_keys($final_tab)) ;
			$limit_fields[] = 'form' ;
		}
		$limit_fields = array_unique($limit_fields) ;
		return parent::save_values_to_db( $limit_fields ) ;
	}
	
	// on fait exactement la même chose pour les créations de lignes
	// il faudrait toujours définir limit_fields, car actuellement la sérialisation ne modifie pas ce tableau
	public function new_values_into_db( $limit_fields=array() ) {
		$final_tab = array() ;
		$field_list = static::get_additional_fieldnames() ;
		foreach( $field_list as $a_fieldname) {
			$final_tab[$a_fieldname] = $this->fields[$a_fieldname]->get_value_sql() ;
		}
		// print "<pre>" ; print_r($final_tab) ; print "\n" ; print serialize( $final_tab ) ; print "</pre>" ; 
		$this->fields['form']->set_value_from_reliable_source( serialize( $final_tab ) ) ;
		if( count($limit_fields)>0) {
			// alors on retire les champs à sérialiser et on ajoute "form"
			$limit_fields = array_diff( $limit_fields, array_keys($final_tab)) ;
			$limit_fields[] = 'form' ;
		}
		$limit_fields = array_unique($limit_fields) ;
		return parent::new_values_into_db( $limit_fields ) ;
	}

	
	// Partant du principe qu'il s'agit ici toujours de BDD -> PHP, alors il faut désérialiser "form" vers "form:*"
	public function fill_from_reliable_array(array $assoc_array, array $limit_fields=array() )  {
		// $assoc_array est par exemple un résultat de PDO, qui contient donc une représentation des champs de la BDD.
		// Il contient donc 'form' mais pas les 'form:*'.
		// Il n'est pas nécessaire de spécifier les valeurs par défaut.
		if( isset($assoc_array['form']) AND $assoc_array['form']!='' ) {
			$tab = unserialize($assoc_array['form']) ;
			$additionnal_fieldnames = static::get_additional_fieldnames() ;
			foreach( $additionnal_fieldnames as $a_field ) {
				if( isset($tab[$a_field]))	$assoc_array[$a_field]=$tab[$a_field] ;
			}
		}
		// le filtre limit_fields n'étant appliqué que plus tard, inutile de le modifier. C'est cohérent ainsi.
		parent::fill_from_reliable_array($assoc_array, $limit_fields ) ;
	}
	
	// ici on module simplement cette méthode pour permettre la suppression du champ "form" si rien n'est spécifié
	public function display_html_tab($complet=TRUE, $limit_fields=array(), $avec_soustitres=FALSE ) {
		if( count($limit_fields)==0) {
			$limit_fields = array_diff(array_keys($this->fields), array('form')) ; 
		}
		return parent::display_html_tab($complet, $limit_fields, $avec_soustitres ) ;
	}

}	





?>